"""Singular assembly."""

import numpy as _np

from bempp_cl.api.assembly import assembler as _assembler
from bempp_cl.api.integration import duffy_galerkin as _duffy_galerkin

import collections as _collections


class SingularAssembler(_assembler.AssemblerBase):
    """Assembler for the singular part of boundary integral operators."""

    def __init__(self, domain, dual_to_range, parameters=None):
        """Instantiate the assembler."""
        super().__init__(domain, dual_to_range, parameters)

    def assemble(self, operator_descriptor, device_interface, precision, *args, **kwargs):
        """Assemble the singular part."""
        from bempp_cl.api.assembly.discrete_boundary_operator import (
            SparseDiscreteBoundaryOperator,
        )
        from bempp_cl.api.utils.helpers import promote_to_double_precision
        from scipy.sparse import coo_matrix, csr_matrix
        from bempp_cl.api.space.space import return_compatible_representation

        domain, dual_to_range = return_compatible_representation(self.domain, self.dual_to_range)
        row_dof_count = dual_to_range.global_dof_count
        col_dof_count = domain.global_dof_count
        row_grid_dofs = dual_to_range.grid_dof_count
        col_grid_dofs = domain.grid_dof_count

        if domain.grid != dual_to_range.grid:
            return SparseDiscreteBoundaryOperator(csr_matrix((row_dof_count, col_dof_count), dtype="float64"))

        trial_local2global = domain.local2global.ravel()
        test_local2global = dual_to_range.local2global.ravel()
        trial_multipliers = domain.local_multipliers.ravel()
        test_multipliers = dual_to_range.local_multipliers.ravel()

        rows, cols, values = assemble_singular_part(
            domain.localised_space,
            dual_to_range.localised_space,
            self.parameters,
            operator_descriptor,
            device_interface,
        )
        global_rows = test_local2global[rows]
        global_cols = trial_local2global[cols]
        global_values = values * trial_multipliers[cols] * test_multipliers[rows]

        if self.parameters.assembly.always_promote_to_double:
            values = promote_to_double_precision(values)

        mat = coo_matrix(
            (global_values, (global_rows, global_cols)),
            shape=(row_grid_dofs, col_grid_dofs),
        ).tocsr()

        if domain.requires_dof_transformation:
            mat = mat @ domain.dof_transformation

        if dual_to_range.requires_dof_transformation:
            mat = dual_to_range.dof_transformation.T @ mat

        return SparseDiscreteBoundaryOperator(mat)


def assemble_singular_part(domain, dual_to_range, parameters, operator_descriptor, device_interface):
    """Actually assemble the Numba kernel."""
    from bempp_cl.api.utils.helpers import get_type
    from bempp_cl.core.dispatcher import singular_assembler_dispatcher
    import bempp_cl.api

    precision = operator_descriptor.precision
    kernel_options = operator_descriptor.options
    is_complex = operator_descriptor.is_complex

    grid = domain.grid
    order = parameters.quadrature.singular

    rule = _SingularQuadratureRuleInterfaceGalerkin(grid, order, dual_to_range.support, domain.support)

    number_of_test_shape_functions = dual_to_range.number_of_shape_functions
    number_of_trial_shape_functions = domain.number_of_shape_functions

    [
        test_points,
        trial_points,
        quad_weights,
        test_elements,
        trial_elements,
        test_offsets,
        trial_offsets,
        weights_offsets,
        number_of_quad_points,
    ] = rule.get_arrays()

    if is_complex:
        result_type = get_type(precision).complex
    else:
        result_type = get_type(precision).real

    result = _np.zeros(
        number_of_test_shape_functions * number_of_trial_shape_functions * len(test_elements),
        dtype=result_type,
    )

    with bempp_cl.api.Timer(message=(f"Singular assembler:{operator_descriptor.identifier}:{device_interface}")):
        singular_assembler_dispatcher(
            device_interface,
            operator_descriptor,
            grid,
            domain,
            dual_to_range,
            test_points,
            trial_points,
            quad_weights,
            test_elements,
            trial_elements,
            test_offsets,
            trial_offsets,
            weights_offsets,
            number_of_quad_points,
            kernel_options,
            result,
        )

    irange = _np.arange(number_of_test_shape_functions)
    jrange = _np.arange(number_of_trial_shape_functions)

    i_ind = _np.tile(_np.repeat(irange, number_of_trial_shape_functions), len(rule.trial_indices)) + _np.repeat(
        rule.test_indices * number_of_test_shape_functions,
        number_of_test_shape_functions * number_of_trial_shape_functions,
    )

    j_ind = _np.tile(_np.tile(jrange, number_of_test_shape_functions), len(rule.trial_indices)) + _np.repeat(
        rule.trial_indices * number_of_trial_shape_functions,
        number_of_test_shape_functions * number_of_trial_shape_functions,
    )

    return (i_ind, j_ind, result)


_SingularQuadratureRule = _collections.namedtuple("_QuadratureRule", "test_points trial_points weights")


class _SingularQuadratureRuleInterfaceGalerkin(object):
    """Interface for a singular quadrature rule."""

    def __init__(self, grid, order, test_support, trial_support):
        """Initialize singular quadrature rule."""
        self._grid = grid
        self._order = order
        self._test_indices = None
        self._trial_indices = None

        self._coincident_rule = _SingularQuadratureRule(*_duffy_galerkin.rule(order, "coincident"))

        self._edge_adjacent_rule = _SingularQuadratureRule(*_duffy_galerkin.rule(order, "edge_adjacent"))

        self._vertex_adjacent_rule = _SingularQuadratureRule(*_duffy_galerkin.rule(order, "vertex_adjacent"))

        # Iterate through the singular pairs and only add those that are
        # in the support of the space.

        self._index_count = {}

        self._coincident_indices = _np.flatnonzero(test_support * trial_support)
        self._index_count["coincident"] = len(self._coincident_indices)

        # test_support and trial_support are boolean arrays.
        # * operation corresponds to and op between the arrays.

        edge_adjacent_pairs = _np.flatnonzero(
            test_support[grid.edge_adjacency[0, :]] * trial_support[grid.edge_adjacency[1, :]]
        )

        self._edge_adjacency = grid.edge_adjacency[:, edge_adjacent_pairs]

        vertex_adjacent_pairs = _np.flatnonzero(
            test_support[grid.vertex_adjacency[0, :]] * trial_support[grid.vertex_adjacency[1, :]]
        )

        self._vertex_adjacency = grid.vertex_adjacency[:, vertex_adjacent_pairs]

        self._index_count["edge_adjacent"] = self._edge_adjacency.shape[1]
        self._index_count["vertex_adjacent"] = self._vertex_adjacency.shape[1]

        self._index_count["all"] = (
            self._index_count["coincident"] + self._index_count["edge_adjacent"] + self._index_count["vertex_adjacent"]
        )

    @property
    def order(self):
        """Return the order."""
        return self._order

    @property
    def coincident_rule(self):
        """Return coincident rule."""
        return self._coincident_rule

    @property
    def edge_adjacent_rule(self):
        """Return edge adjacent rule."""
        return self._edge_adjacent_rule

    @property
    def vertex_adjacent_rule(self):
        """Return vertex adjacent rule."""
        return self._vertex_adjacent_rule

    @property
    def grid(self):
        """Return the grid."""
        return self._grid

    @property
    def edge_adjacency(self):
        """Return the grid edge adjacency information."""
        return self._edge_adjacency

    @property
    def vertex_adjacency(self):
        """Return vertex adjacency."""
        return self._vertex_adjacency

    @property
    def number_of_elements(self):
        """Return the number of elements of the underlying grid."""
        return self.grid.number_of_elements

    @property
    def index_count(self):
        """Return the index count."""
        return self._index_count

    @property
    def test_indices(self):
        """Return the test indicies of all singular contributions."""
        return self._test_indices

    @property
    def trial_indices(self):
        """Return the trial indicies of all singular contributions."""
        return self._trial_indices

    def number_of_points(self, adjacency):
        """Return the number of quadrature points for given adjacency."""
        return _duffy_galerkin.number_of_quadrature_points(self.order, adjacency)

    def get_arrays(self):
        """Return the arrays."""
        test_indices, trial_indices = self._vectorize_indices()
        test_points, trial_points = self._vectorize_points()
        weights = self._vectorize_weights()
        test_offsets, trial_offsets, weights_offsets = self._vectorize_offsets()
        number_of_quad_points = self._get_number_of_quad_points()

        self._test_indices = test_indices
        self._trial_indices = trial_indices

        arrays = [
            test_points,
            trial_points,
            weights,
            test_indices,
            trial_indices,
            test_offsets,
            trial_offsets,
            weights_offsets,
            number_of_quad_points,
        ]

        return arrays

    def _collect_remapped_quad_points_for_edge_adjacent_rule(self, quad_points):
        """
        Remap quad points for edge adjacent quadrature rules.

        Given a 2xN array of quadrature points, return all possible
        combinations of remapped rules, according to the following
        order
        0: edge (index 0, 1)
        1: edge (index 1, 0)
        2: edge (index 1, 2)
        3: edge (index 2, 1)
        4: edge (index 0, 2)
        5: edge (index 2, 0)
        """
        return _np.hstack(
            [
                _duffy_galerkin.remap_points_shared_edge(quad_points, 0, 1),
                _duffy_galerkin.remap_points_shared_edge(quad_points, 1, 0),
                _duffy_galerkin.remap_points_shared_edge(quad_points, 1, 2),
                _duffy_galerkin.remap_points_shared_edge(quad_points, 2, 1),
                _duffy_galerkin.remap_points_shared_edge(quad_points, 0, 2),
                _duffy_galerkin.remap_points_shared_edge(quad_points, 2, 0),
            ]
        )

    def _collect_remapped_quad_points_for_vertex_adjacent_rule(self, quad_points):
        """
        Remap quad points for vertex adjacent quadrature rules.

        Given a 2xN array of quadrature points, return all possible
        combinations of remapped rules, according to the following
        order
        0: vertex (index 0)
        1: vertex (index 1)
        2: vertex (index 2)
        """
        return _np.hstack(
            [
                _duffy_galerkin.remap_points_shared_vertex(quad_points, 0),
                _duffy_galerkin.remap_points_shared_vertex(quad_points, 1),
                _duffy_galerkin.remap_points_shared_vertex(quad_points, 2),
            ]
        )

    def _compute_edge_offsets(self):
        """Compute offsets for the edge based rule."""
        ncoincident = self.number_of_points("coincident")
        nedge_adjacent = self.number_of_points("edge_adjacent")

        # Offset values is a 3 x 3 matrix such that
        # the value (i, j) is the offset of the rule
        # associted with the (i, j) remap case,
        # where i and j are 0, 1, or 2. The diagonal
        # elements are not needed, so just set to -1.

        offset_values = _np.array([[-1, 0, 4], [1, -1, 2], [5, 3, -1]])

        edge_offsets = ncoincident + nedge_adjacent * offset_values

        return edge_offsets

    def _compute_vertex_offsets(self):
        """Compute offsets for the vertex based rules."""
        ncoincident = self.number_of_points("coincident")
        nedge_adjacent = self.number_of_points("edge_adjacent")
        nvertex_adjacent = self.number_of_points("vertex_adjacent")

        vertex_offsets = ncoincident + 6 * nedge_adjacent + nvertex_adjacent * _np.arange(3, dtype="uint32")

        return vertex_offsets

    def _vectorize_indices(self):
        """Return vector of test and trial indices for sing. integration."""
        test_indices = _np.empty(self.index_count["all"], dtype="uint32")
        trial_indices = _np.empty(self.index_count["all"], dtype="uint32")

        for array, index in zip([test_indices, trial_indices], [0, 1]):
            count = self._index_count["coincident"]
            array[:count] = self._coincident_indices

            array[count : (count + self.index_count["edge_adjacent"])] = self.edge_adjacency[index, :]

            count += self.index_count["edge_adjacent"]

            array[count:] = self.vertex_adjacency[index, :]

        return test_indices, trial_indices

    def _get_number_of_quad_points(self):
        """Compute an array of local numbers of integration points."""
        n = self.index_count["all"]
        number_of_quad_points = _np.empty(self.index_count["all"], dtype="uint32")

        number_of_quad_points[: self.index_count["coincident"]] = self.number_of_points("coincident")
        number_of_quad_points[
            self.index_count["coincident"] : (self.index_count["coincident"] + self.index_count["edge_adjacent"])
        ] = self.number_of_points("edge_adjacent")
        number_of_quad_points[n - self.index_count["vertex_adjacent"] :] = self.number_of_points("vertex_adjacent")

        return number_of_quad_points

    def _vectorize_points(self):
        """Return an array of all quadrature points for all different rules."""
        test_points = _np.hstack(
            [
                self.coincident_rule.test_points,
                self._collect_remapped_quad_points_for_edge_adjacent_rule(self.edge_adjacent_rule.test_points),
                self._collect_remapped_quad_points_for_vertex_adjacent_rule(self.vertex_adjacent_rule.test_points),
            ]
        )

        trial_points = _np.hstack(
            [
                self.coincident_rule.trial_points,
                self._collect_remapped_quad_points_for_edge_adjacent_rule(self.edge_adjacent_rule.trial_points),
                self._collect_remapped_quad_points_for_vertex_adjacent_rule(self.vertex_adjacent_rule.trial_points),
            ]
        )

        return test_points, trial_points

    def _vectorize_weights(self):
        """Vectorize the quadrature weights."""
        weights = _np.hstack(
            [
                self.coincident_rule.weights,
                self.edge_adjacent_rule.weights,
                self.vertex_adjacent_rule.weights,
            ]
        )

        return weights

    def _vectorize_offsets(self):
        """Vectorize the offsets."""
        edge_offsets = self._compute_edge_offsets()
        vertex_offsets = self._compute_vertex_offsets()

        test_offsets = _np.empty(self.index_count["all"], dtype="uint32")
        trial_offsets = _np.empty(self.index_count["all"], dtype="uint32")
        weights_offsets = _np.empty(self.index_count["all"], dtype="uint32")

        test_offsets[: self.index_count["coincident"]] = 0

        test_offsets[
            self.index_count["coincident"] : (self.index_count["coincident"] + self.index_count["edge_adjacent"])
        ] = edge_offsets[self.edge_adjacency[2, :], self.edge_adjacency[3, :]]

        test_offsets[(self.index_count["coincident"] + self.index_count["edge_adjacent"]) :] = vertex_offsets[
            self.vertex_adjacency[2, :]
        ]

        trial_offsets[: self.index_count["coincident"]] = _np.zeros(self.index_count["coincident"])

        trial_offsets[
            self.index_count["coincident"] : (self.index_count["coincident"] + self.index_count["edge_adjacent"])
        ] = edge_offsets[self.edge_adjacency[4, :], self.edge_adjacency[5, :]]

        trial_offsets[(self.index_count["coincident"] + self._index_count["edge_adjacent"]) :] = vertex_offsets[
            self.vertex_adjacency[3, :]
        ]

        weights_offsets[: self.index_count["coincident"]] = 0
        weights_offsets[
            self.index_count["coincident"] : (self.index_count["coincident"] + self.index_count["edge_adjacent"])
        ] = self.number_of_points("coincident")
        weights_offsets[(self.index_count["coincident"] + self.index_count["edge_adjacent"]) :] = self.number_of_points(
            "coincident"
        ) + self.number_of_points("edge_adjacent")

        return test_offsets, trial_offsets, weights_offsets



import numpy as np
import scipy as sp
from scipy.interpolate import interp1d

c = 343
rho = 1.22


## Thiele/Small param estimator
def getThieleSmallParam(impedance_file, Bl, peak_max_bound=3000, skiprows=1, delimiter=",", freqCol=0, HCol=1, PCol=2, usecols=None):
    """
    Get the Thiele/Small param from impedance curve and force factor Bl.

    Parameters:
    - impedance_file: .txt file
        containing the impedance as H*exp(1j*Phi) -> H column 2 and Phi column 3 (freq being 1)
    - Bl: int, float
        force factor of the motor

    Return:
    - Optimized Thiele/Small Parameters

    Note:
    Me just following matlab code; though the translation Matlab -> Python didn't work
    so I used curve_fit instead of fmin (I let Scipy do the error reduction because me dum dum)
    """
    from electroacPy.general import findInArray, parallel
    
    # 1. Load impedance data
    data = np.loadtxt(impedance_file, skiprows=skiprows, delimiter=delimiter, usecols=usecols)

    freq = data[:, freqCol]
    try:
        Ze_meas = data[:, HCol] * np.exp(1j * data[:, PCol]) # H*e^(j*Phi)
    except:
        Ze_meas = data[:, HCol]

    # 2. Get some visible values
    ind_bound, _ = findInArray(freq, peak_max_bound)
    Zmax = np.max(np.abs(Ze_meas[:ind_bound]))
    ind_fc, _ = findInArray(np.abs(Ze_meas), Zmax)
    fc = freq[ind_fc]

    # 3. Reduce the size of frequency range if it goes above 20 kHz
    ind_fmax, fmax = findInArray(freq, 20e3)
    freq = freq[:ind_fmax + 1]

    # 4. Start Values
    Re = np.min(np.abs(Ze_meas))
    Le = 1e-5
    Res = Zmax - Re
    R2, L2 = 0.5, 2e-4
    R3, L3 = 0.25, 1e-4
    Wc, Qmc = 2 * np.pi * fc, getQMC(freq, Ze_meas)
    x0 = [Re, Le, Res, R2, L2, R3, L3, Wc, Qmc] # initial parameters

    # 5. Merci Scipy
    popt, pcov = sp.optimize.curve_fit(sealed_impedance_calc,
                                       freq, np.abs(Ze_meas), p0=x0, maxfev=16000)
    Re, Le = popt[0], popt[1]
    Res = popt[2]
    R2, L2 = popt[3], popt[4]
    R3, L3 = popt[5], popt[6]
    Wc, Qmc = popt[7], popt[8]
    fc = Wc / 2 / np.pi
    sc = 1j * Wc

    # -----------------
    # 6. Let's finalize
    Mmc = Qmc / Wc * Bl**2/Res
    Cmc = 1 / Wc**2 / Mmc

    # impedance at resonance
    Zfc = np.abs(Re + sc*Le + parallel(sc*L2, R2) + parallel(sc*L3, R3))

    # Q factors
    Qec = np.sqrt(Mmc/Cmc) * (Zfc / Bl**2)
    Qtc = Qmc*Qec/(Qmc+Qec)

    # out param
    params = {}
    params['Re']  = Re
    params['Le']  = Le
    params['Res'] = Res
    params['R2']  = R2
    params['L2']  = L2
    params['R3']  = R3
    params['L3']  = L3
    params['Wc']  = Wc
    params['fc']  = fc
    params['Qmc'] = Qmc
    params['Cmc'] = Cmc
    params['Mmc'] = Mmc
    params['Zfc'] = Zfc
    params['Qec'] = Qec
    params['Qtc'] = Qtc
    return params

def getQMC(freq, Z, stepf=1e-3):
    """
    Find the quality factor of the impedance peak of a driver in a sealed box.
    :param freq:
    :param Z:
    :param stepf:
    :return:
    """
    from electroacPy.general import findInArray

    # increase a bit the resolution (doesn't change curve's shape)
    finterp = np.arange(freq[0], freq[-1] + stepf, stepf)
    Zinterp = np.interp(finterp, freq, np.abs(Z))

    # get index of Zmax (index of f0) and Zmax
    ind_f0, Zmax = findInArray(np.abs(Zinterp), np.max(np.abs(Zinterp)))
    Ef0 = finterp[ind_f0]

    ind_fmin, Zfmin = findInArray(np.abs(Zinterp[:ind_f0]),
                                      np.max(np.abs(Zinterp)) * 0.707)  # get -3 dB value before resonance
    ind_fmax, Zfmax = findInArray(np.abs(Zinterp[ind_f0:]), np.max(np.abs(Zinterp)) * 0.707)
    range = [finterp[ind_fmin], finterp[ind_f0 + ind_fmax]]
    Qs = finterp[ind_f0] / (range[1] - range[0])
    return Qs


def sealed_impedance_calc(freq, Re, Le, Res, R2, L2, R3, L3, Wc, Qmc):
    """
    sealed impedance calculator, should be use within the curve_fit() algorithm
    :param freq:
    :param Re:
    :param Le:
    :param Res:
    :param R2:
    :param L2:
    :param R3:
    :param L3:
    :param Wc:
    :param Qmc:
    :return:
    """
    from electroacPy.general import parallel

    s = 1j * 2 * np.pi * freq

    # mechanical impedance
    fncZm = Res / (1 + Qmc * (s / Wc + Wc / s))

    # electrical impedance
    fncZe = Re + s * Le + parallel(s * L2, R2) + parallel(s * L3, R3)

    # total impedance
    fncZ = fncZe + fncZm
    return np.abs(fncZ)

def Zes_impedance(freq, Bl, Rms, Mms, Cms):
    s = 2j * np.pi * freq
    Zems = Bl ** 2 / (Rms + s*Mms + 1/s/Cms)
    return np.abs(Zems)

def estimate_pmec(Bl, Qms, Res, fs):
    Mms = Qms * Bl**2 / 2 / np.pi / fs / Res
    Rms = Bl**2 / Res
    Cms = Res**2 / Bl**4 * Mms / Qms**2
    return Mms, Rms, Cms

## Room Acoustics
def art60(rt60, surface, volume, f=1000, c=343):
    """
    Compute the absorption coefficient based on the given reverberation time (RT60),
    total surface area of absorbing materials in the room (surface), and the volume
    of the room (volume). Compute for f=1000 Hz.

    Parameters:
    - rt60 (float): Reverberation time in seconds.
    - surface (float): Total surface area of absorbing materials in square meters.
    - volume (float): Volume of the room in cubic meters.

    Returns:
    float: Absorption coefficient calculated based on the provided inputs.

    Example:
    >>> art60(1.5, 50, 100)
    """

    om = 2*np.pi*f
    k = om*c
    a = k * volume / rt60 / surface
    return a


def a2z(absorptionCoeff, c=343, rho=1.22):
    Z_air = rho * c
    Zn = Z_air * (1-np.sqrt(1-absorptionCoeff)) / (1+np.sqrt(1-absorptionCoeff))
    return Zn


def rectangularRoomModes(Lx, Ly, Lz, N):
    """
    Compute the Nth room modes of a rectangular room given its dimensions.

    Parameters:
    - Lx (float): Length of the room in the x-direction.
    - Ly (float): Width of the room in the y-direction.
    - Lz (float): Height of the room in the z-direction.
    - N (int): Amount of modes to compute (N^3).

    Returns:
    - modes (list): A list containing the Nth room modes in the format (fx, fy, fz),
                   where fx, fy, and fz are the frequencies along the x, y, and z directions, respectively.

    Note:
    - Room modes represent the resonant frequencies at which standing waves can exist in the room.
    - The mode numbers (N) correspond to the number of half-wavelengths along each dimension.
    - The formula for the Nth room mode frequencies is given by:
        f[x, y, z] = 343/2 * ((nx/Lx)^2 + (ny/Ly)^2 + (nz/Lz)^2)^0.5
    """

    f = np.zeros([N, N, N])
    for nx in range(N):
        for ny in range(N):
            for nz in range(N):
                f[nx, ny, nz] = 343/2 * np.sqrt((nx/Lx)**2+(ny/Ly)**2+(nz/Lz)**2)
    return f


def interpolate_Impedance(impedanceData, freq_in, freq_out):
    """
    Interpolate impedance data from freq_in array to freq_out array.

    Parameters:
    - impedanceData: numpy array, impedance values corresponding to freq_in.
    - freq_in: numpy array, input frequency array.
    - freq_out: numpy array, output frequency array for interpolation.

    Returns:
    - interpolatedImpedance: numpy array, interpolated impedance values corresponding to freq_out.
    """

    # Create an interpolation function
    interpolation_function = interp1d(freq_in, impedanceData, kind='linear', fill_value="extrapolate")

    # Use the interpolation function to compute impedance values at freq_out
    interpolatedImpedance = interpolation_function(freq_out)

    return interpolatedImpedance


#%% Export tools
def export_directivity(folder_name, file_prefix,
                       frequency_array, angle_array, pmic_array, frd=False):
    """
    Export a directivity evaluation extracted from a loudspeakerSystem object.
    Can be imported in other software, such as VituixCAD or VACS.

    Parameters
    ----------
    folder_name : str
        export path.
    frequency_array : numpy array
        frequency range of length Nfft.
    angle_array : numpy array
        angular position of microphones
    pmic_array : numpy array
        acoustic pressure (complex) of shape [Nfft, N_angles].

    Returns
    -------
    folder with *.txt file containing frequency, SPL and phase of extracted 
    directivity.

    """
    import os
    from electroacPy.general.gain import SPL
    
    # Create the folder if it doesn't exist
    if not os.path.exists(folder_name):
        os.makedirs(folder_name)
    
    if frd is False:
        extension = "txt"
    elif frd is True:
        extension = "frd"
    
    # Loop through each microphone (angle)
    for i, angle in enumerate(angle_array):
        # Compute SPL values for the current microphone
        spl_values = SPL(pmic_array[:, i])
        
        # Compute phase (in degrees) using np.angle and np.rad2deg
        phase_values = np.rad2deg(np.angle(pmic_array[:, i]))
        
        # Prepare the data to save: frequency, SPL, and phase
        data_to_save = np.column_stack((frequency_array, spl_values, phase_values))
        
        # File name format: folder_name_angle.txt
        file_name = os.path.join(folder_name, f"{file_prefix}_{angle}.{extension}")
        
        # Save the data to the text file using np.savetxt with the correct header
        np.savetxt(file_name, data_to_save, header="frequency  SPL  phase", fmt="%.3e")
        
        # print(f"Saved data to {file_name}")


"""Main interface class to ExaFMM."""

import numpy as _np
import atexit as _atexit

FMM_TMP_DIR = None


@_atexit.register
def cleanup_fmm_tmp():
    """Clean up the FMM tmp directory."""
    from pathlib import Path

    try:
        if FMM_TMP_DIR is not None:
            for tmp_file in Path(FMM_TMP_DIR).glob("*.tmp"):
                tmp_file.unlink()
    except:  # noqa: E722
        print("Could not delete FMM temporary files.")


class ExafmmInterface(object):
    """Interface to Exafmm."""

    def __init__(
        self,
        source_points,
        target_points,
        mode,
        wavenumber=None,
        depth=4,
        expansion_order=5,
        ncrit=400,
        precision="double",
        singular_correction=None,
    ):
        """Instantiate an Exafmm session."""
        import bempp_cl.api
        import os
        from bempp_cl.api.utils.helpers import create_unique_id

        global FMM_TMP_DIR

        if FMM_TMP_DIR is None:
            FMM_TMP_DIR = os.path.join(os.getcwd(), ".exafmm")
            if not os.path.isdir(FMM_TMP_DIR):
                try:
                    os.mkdir(FMM_TMP_DIR)
                except:  # noqa: E722
                    raise FileExistsError(f"A file with the name {FMM_TMP_DIR} exists. Please delete it.")

        for _ in range(10):
            tmp_name = create_unique_id() + ".tmp"
            fname = os.path.join(FMM_TMP_DIR, tmp_name)
            if not os.path.exists(fname):
                break
        else:
            raise FileExistsError("Could not create temporary filename for Exafmm.")

        self._fname = fname
        self._singular_correction = singular_correction

        self._source_points = source_points
        self._target_points = target_points
        self._mode = mode

        if mode == "laplace":
            self._kernel_parameters = _np.array([], dtype="float64")
        elif mode == "helmholtz":
            self._kernel_parameters = _np.array([_np.real(wavenumber), _np.imag(wavenumber)], dtype="float64")
        elif mode == "modified_helmholtz":
            self._kernel_parameters = _np.array([wavenumber], dtype="float64")

        with bempp_cl.api.Timer(message="Initialising Exafmm."):
            if mode == "laplace":
                import exafmm.laplace

                self._module = exafmm.laplace

                sources = exafmm.laplace.init_sources(source_points, _np.zeros(len(source_points), dtype=_np.float64))

                targets = exafmm.laplace.init_targets(target_points)

                self._fmm = exafmm.laplace.LaplaceFmm(expansion_order, ncrit, filename=fname)
                self._tree = exafmm.laplace.setup(sources, targets, self._fmm)

            elif mode == "helmholtz":
                import exafmm.helmholtz

                self._module = exafmm.helmholtz

                sources = exafmm.helmholtz.init_sources(source_points, _np.zeros(len(source_points), dtype=_np.float64))

                targets = exafmm.helmholtz.init_targets(target_points)

                self._fmm = exafmm.helmholtz.HelmholtzFmm(expansion_order, ncrit, wavenumber, filename=fname)
                self._tree = exafmm.helmholtz.setup(sources, targets, self._fmm)

            elif mode == "modified_helmholtz":
                import exafmm.modified_helmholtz

                self._module = exafmm.modified_helmholtz

                sources = exafmm.modified_helmholtz.init_sources(
                    source_points, _np.zeros(len(source_points), dtype=_np.float64)
                )

                targets = exafmm.modified_helmholtz.init_targets(target_points)

                self._fmm = exafmm.modified_helmholtz.ModifiedHelmholtzFmm(
                    expansion_order, ncrit, wavenumber, filename=fname
                )
                self._tree = exafmm.modified_helmholtz.setup(sources, targets, self._fmm)

    @property
    def number_of_source_points(self):
        """Return number of source points."""
        return len(self._source_points)

    @property
    def number_of_target_points(self):
        """Return number of target points."""
        return len(self._target_points)

    def evaluate(self, vec, apply_singular_correction=True):
        """Evalute the Fmm."""
        import bempp_cl.api
        from bempp_cl.api.fmm.helpers import debug_fmm

        with bempp_cl.api.Timer(message="Evaluating Fmm."):
            self._module.update_charges(self._tree, vec)
            self._module.clear_values(self._tree)

            with bempp_cl.api.Timer(message="Calling ExaFMM."):
                if bempp_cl.api.GLOBAL_PARAMETERS.fmm.dense_evaluation:
                    from bempp_cl.api.fmm.helpers import dense_interaction_evaluator

                    result = dense_interaction_evaluator(
                        self._target_points,
                        self._source_points,
                        vec,
                        self._mode,
                        self._kernel_parameters,
                    )
                else:
                    result = self._module.evaluate(self._tree, self._fmm)
                if bempp_cl.api.GLOBAL_PARAMETERS.fmm.debug:
                    debug_fmm(
                        self._target_points,
                        self._source_points,
                        vec,
                        self._mode,
                        self._kernel_parameters,
                        result,
                    )

            if apply_singular_correction and self._singular_correction is not None:
                result -= (self._singular_correction @ vec).reshape([-1, 4])

            return result

    def as_matrix(self):
        """Return matrix representation of Fmm."""
        import numpy as np

        ident = np.identity(self.number_of_source_points)

        res = np.zeros(
            (self.number_of_target_points, self.number_of_source_points),
            dtype="float64",
        )

        for index in range(self.number_of_source_points):
            res[:, index] = self.evaluate(ident[:, index])[:, 0]

        return res

    @classmethod
    def from_grid(
        cls,
        source_grid,
        mode,
        wavenumber=None,
        target_grid=None,
        precision="double",
        parameters=None,
        device_interface=None,
    ):
        """
        Initialise an Exafmm instance from a given source and target grid.

        Parameters
        ----------
        source_grid : Grid object
            Grid for the source points.
        mode: string
            Fmm mode. One of 'laplace', 'helmholtz', or 'modified_helmholtz'
        wavenumber : real number
            For Helmholtz or modified Helmholtz the wavenumber.
        target_grid : Grid object
            An optional target grid. If not provided the source and target
            grid are assumed to be identical.
        precision : string
            Either 'single' or 'double'. Currently, the Fmm is always
            executed in double precision.
        parameters  :   object
            A bempp parameters object. If not provided, the GLOBAL_PARAMETERS will be used.
        device_interface : string
            Either 'numba' or 'opencl'. If not provided, the DEFAULT_DEVICE_INTERFACE
            will be used for the calculation of local interactions.
        """
        import bempp_cl.api
        from bempp_cl.api.integration.triangle_gauss import rule
        from bempp_cl.api.fmm.helpers import get_local_interaction_operator
        import numpy as np

        parameters = bempp_cl.api.assign_parameters(parameters)

        quadrature_order = parameters.quadrature.regular

        local_points, weights = rule(quadrature_order)

        if target_grid is None:
            target_grid = source_grid

        source_points = source_grid.map_to_point_cloud(quadrature_order, precision=precision)

        if target_grid != source_grid:
            target_points = target_grid.map_to_point_cloud(quadrature_order, precision=precision)
        else:
            target_points = source_points

        singular_correction = None

        if target_grid == source_grid:
            # Require singular correction terms.

            if mode == "laplace":
                singular_correction = get_local_interaction_operator(
                    source_grid,
                    local_points,
                    "laplace",
                    np.array([], dtype="float64"),
                    precision,
                    False,
                    device_interface,
                )
            elif mode == "helmholtz":
                singular_correction = get_local_interaction_operator(
                    source_grid,
                    local_points,
                    "helmholtz",
                    np.array([_np.real(wavenumber), _np.imag(wavenumber)], dtype="float64"),
                    precision,
                    True,
                    device_interface,
                )
            elif mode == "modified_helmholtz":
                singular_correction = get_local_interaction_operator(
                    source_grid,
                    local_points,
                    "modified_helmholtz",
                    np.array([wavenumber], dtype="float64"),
                    precision,
                    False,
                    device_interface,
                )
        return cls(
            source_points,
            target_points,
            mode,
            wavenumber=wavenumber,
            depth=parameters.fmm.depth,
            expansion_order=parameters.fmm.expansion_order,
            ncrit=parameters.fmm.ncrit,
            precision=precision,
            singular_correction=singular_correction,
        )



"""Implementation of an octree in Python."""

import numpy as _np
import numba as _numba


@_numba.experimental.jitclass(
    [
        ("_lbound", _numba.float64[:]),
        ("_ubound", _numba.float64[:]),
        ("_maximum_level", _numba.int32),
        ("_diameter", _numba.float64[:]),
        ("_vertices", _numba.float64[:, :]),
        ("_sorted_indices", _numba.uint32[:]),
        ("_leaf_nodes", _numba.uint32[:]),
        ("_leaf_nodes_index_ptr", _numba.uint32[:]),
        ("_level_nodes", _numba.uint32[:]),
        ("_level_nodes_index_ptr", _numba.uint32[:]),
        ("_near_field_nodes", _numba.int32[:]),
    ]
)
class Octree(object):
    """Data structure for handling Octrees."""

    def __init__(self, lbound, ubound, maximum_level, vertices):
        """
        Initialize an Octree.

        Parameters
        ----------
        lbound : np.ndarray
            Numpy array of size (3, ) that specifies the lower
            bound of the Octree.
        ubound : np.ndarray
            Numpy array of size (3, ) that specifies the upper
            bound of the Octree.
        maximum_level : integer
            The maximum level of the Octree.
        vertices : np.ndarray
            An (3, N) float64 array of N vertices
        """
        self._lbound = lbound
        self._ubound = ubound
        self._maximum_level = maximum_level
        self._diameter = ubound - lbound
        self._vertices = vertices
        self._assign_nodes(vertices)
        self._compute_nearfields()

    @property
    def diameter(self):
        """Return diameter of the Octree in each dimension."""
        return self._diameter

    @property
    def lower_bound(self):
        """Return lower bound of Octree in each dimension."""
        return self._lbound

    @property
    def upper_bound(self):
        """Return upper bound of Octree in each dimension."""
        return self._ubound

    @property
    def maximum_level(self):
        """Return the maximum level."""
        return self._maximum_level

    @property
    def vertices(self):
        """Return the vertices."""
        return self._vertices

    @property
    def non_empty_leaf_nodes(self):
        """Return the non-empty leaf nodes."""
        return self._leaf_nodes

    @property
    def sorted_indices(self):
        """Return the indices sorted by leaf node."""
        return self._sorted_indices

    @property
    def leaf_nodes_ptr(self):
        """
        Return the index pointers for the leaf node elements.

        Returns an array index_ptr, such that the indices
        from the jth non-empty leaf node can be obtained
        by self.sorted_indices[index_ptr[j]:index_ptr[j+1]]
        and the associated node index through
        self._non_empty_leaf_nodes[j].
        """
        return self._leaf_nodes_index_ptr

    @property
    def non_empty_nodes_by_level(self):
        """Return the non-empty nodes by level."""
        return self._level_nodes

    @property
    def non_empty_nodes_ptr(self):
        """
        Return the index pointers for non-empty nodes by level.

        Returns an array index_ptr, such that the non-empty nodes
        of the jth level are given by
        self.non_empty_nodes_by_level[index_ptr[j]:index_ptr[j+1]].
        """
        return self._level_nodes_index_ptr

    @property
    def near_field_nodes(self):
        """Return near field nodes."""
        return self._near_field_nodes

    @property
    def near_field_nodes_ptr(self):
        """
        Return an index ptr to the near field nodes.

        Returns an array index_ptr, such that
        self.near_field_nodes[index_ptr[j]:index_ptr[j+1]]
        contains the near field nodes for all nodes in level j.
        There are 27 entries for each near field node. The
        sequence of nodes is the same as for the array
        non_empty_nodes_by_level.
        """
        return _np.uint32(27) * self._level_nodes_index_ptr

    def parent(self, node_index):
        """Return the parent index of a node."""
        return node_index >> 3

    def children(self, node_index):
        """Return an iterator over the child indices."""
        first = node_index << 3
        last = 7 + (node_index << 3)
        return list(range(first, 1 + last))

    def nodes_per_side(self, level):
        """Return number of nodes along each dimension."""
        return 1 << level

    def nodes_per_level(self, level):
        """Return the number of nodes in a given level."""
        return 1 << 3 * level

    def leaf_containing_point(self, point):
        """Return the Morton index of a node containing the point."""
        leaf_size = self.nodes_per_side(self.maximum_level)

        fractions = ((point - self.lower_bound) / self.diameter) * leaf_size

        indices = _np.fmin(_np.fmax(0, fractions).astype(_np.int32), leaf_size - 1)

        return morton(indices)

    def node_bounds(self, morton_index, level):
        """
        Return the lower/upper bound of a node by Morton index.

        The method returns a tuple (lbound, ubound) which define
        the lower and upper corners of a node given by its Morton index.
        """
        indices = _np.array(de_morton(morton_index))
        nnodes_along_dimension = self.nodes_per_side(level)

        node_size = self.diameter / nnodes_along_dimension
        lbound = self.lower_bound + indices * node_size
        ubound = self.lower_bound + (1 + indices) * node_size
        return (lbound, ubound)

    def neighbors(self, node_index, level):
        """Return a list of indices of the neighbors of a node."""
        return _neighbors(node_index, level)

    def node_diameter(self, level):
        """Return node diameter in a given level."""
        return self.diameter / (1.0 * self.nodes_per_side(self.maximum_level))

    def _assign_nodes(self, vertices):
        """Compute leaf-nodes and parents."""
        nvertices = vertices.shape[1]
        node_indices = _np.empty(nvertices, dtype=_np.uint32)

        for index in range(nvertices):
            node_indices[index] = self.leaf_containing_point(vertices[:, index])

        self._sorted_indices = _np.argsort(node_indices).astype(_np.uint32)
        index_ptr = []
        nodes = []

        tmp = -1

        for i, index in enumerate(self._sorted_indices):
            if tmp != node_indices[index]:
                index_ptr.append(i)
                nodes.append(node_indices[index])
                tmp = node_indices[index]

        index_ptr.append(len(self._sorted_indices))
        self._leaf_nodes = _np.array(nodes, dtype=_np.uint32)
        self._leaf_nodes_index_ptr = _np.array(index_ptr, dtype=_np.uint32)

        # Now compute the parents

        self._level_nodes = self._leaf_nodes
        self._level_nodes_index_ptr = _np.zeros(self.maximum_level + 2, _np.uint32)
        nnodes = _np.empty(1 + self._maximum_level, _np.uint32)
        nnodes[self.maximum_level] = len(self._leaf_nodes)
        current_level = self._leaf_nodes
        current_nodes = []

        for index in range(1, self.maximum_level + 1):
            current_nodes.clear()
            for node_index in current_level:
                current_nodes.append(self.parent(node_index))
            current_level = _make_unique(_np.array(current_nodes, dtype=_np.uint32))
            self._level_nodes = _np.concatenate((current_level, self._level_nodes))
            nnodes[self.maximum_level - index] = len(current_level)

        tmp = 0

        for index in range(0, 1 + self.maximum_level):
            self._level_nodes_index_ptr[index] = tmp
            tmp += nnodes[index]
        self._level_nodes_index_ptr[1 + self.maximum_level] = tmp

    def _compute_nearfields(self):
        """
        Compute near fields of all non empty nodes.

        Each node can have at most 27 near field nodes (including
        the node itself). If a near field node does not exist or is empty
        then the value -1 is stored, otherwise the node number.
        """
        self._near_field_nodes = _np.empty(27 * len(self._level_nodes), _np.int32)

        count = 0

        for level_index in range(self.maximum_level + 1):
            sides = 1 << level_index
            level_nodes = self.non_empty_nodes_by_level[
                self.non_empty_nodes_ptr[level_index] : self.non_empty_nodes_ptr[level_index + 1]
            ]
            for node_index in level_nodes:
                ind1, ind2, ind3 = de_morton(node_index)

                for i in range(-1, 2):
                    for j in range(-1, 2):
                        for k in range(-1, 2):
                            if _in_range(ind1 + i, ind2 + j, ind3 + k, sides):
                                morton_index = morton((ind1 + i, ind2 + j, ind3 + k))
                                if _np.any(level_nodes == morton_index):
                                    self._near_field_nodes[count] = morton_index
                                else:
                                    self._near_field_nodes[count] = -1
                            else:
                                self._near_field_nodes[count] = -1
                            count += 1

    def _compute_interaction_list(self):
        """Compute the interaction list for each non empty node."""
        pass


@_numba.njit
def _make_unique(ar):
    """Find unique elements.

    An implementation of Numpy unique for Numba.
    """
    sorted_ar = _np.sort(ar)
    unique_lst = [sorted_ar[0]]
    for elem in sorted_ar:
        if elem != unique_lst[-1]:
            unique_lst.append(elem)
    return _np.array(unique_lst, dtype=ar.dtype)


@_numba.njit(cache=True)
def _in_range(n1, n2, n3, bound):  # pylint: disable=C0103
    """Check if 0 <= n1, n2, n3 < bound."""
    return n1 >= 0 and n1 < bound and n2 >= 0 and n2 < bound and n3 >= 0 and n3 < bound


@_numba.njit(cache=True)
def morton(indices):
    """Encode an integer tuple (i1, i2, i3) via Morton encoding."""
    # pylint: disable=C0103
    x, y, z = indices
    return _dilate(x) | (_dilate(y) << 1) | (_dilate(z) << 2)


@_numba.njit(cache=True)
def de_morton(index):
    """Decode a Morton index."""
    ind1 = _contract(index)
    ind2 = _contract(index >> 1)
    ind3 = _contract(index >> 2)

    return (ind1, ind2, ind3)


@_numba.njit(cache=True)
def _neighbors(node_index, level):
    """Return a list of neighbors of a given node."""
    sides = 1 << level

    ind1, ind2, ind3 = de_morton(node_index)

    result = []

    for i in range(-1, 2):
        for j in range(-1, 2):
            for k in range(-1, 2):
                if i == 0 and j == 0 and k == 0:
                    continue
                if _in_range(ind1 + i, ind2 + j, ind3 + k, sides):
                    result.append(morton((ind1 + i, ind2 + j, ind3 + k)))
    return result


@_numba.njit(cache=True)
def _dilate(number):
    """Dilate an integer for the Morton encoding."""
    number = (number | (number << 16)) & 0x030000FF
    number = (number | (number << 8)) & 0x0300F00F
    number = (number | (number << 4)) & 0x030C30C3
    number = (number | (number << 2)) & 0x09249249

    # Explanation
    # x = ---- ---- ---- ---- ---- --98 7654 3210
    # x = (x | (x << 16)) & 0x030000FF
    # x = ---- --98 ---- ---- ---- ---- 7654 3210
    # x = (x | (x << 8)) & 0x0300F00F
    # x = ---- --98 ---- ---- 7654 ---- ---- 3210
    # x = (x | (x << 4)) & 0x030C30C3
    # x = ---- --98 ---- 76-- --54 ---- 32-- --10
    # x = (x | (x << 2)) & 0x09249249
    # x = ---- 9--8 --7- -6-- 5--4 --3- -2-- 1--0

    return number


@_numba.njit(cache=True)
def _contract(number):
    """Undo dilation."""
    number = number & 0x09249249
    number = (number | (number >> 2)) & 0x030C30C3
    number = (number | (number >> 4)) & 0x0300F00F
    number = (number | (number >> 8)) & 0x030000FF
    number = (number | (number >> 16)) & 0x000003FF

    return number

    # Explanation
    #  x &= 0x09249249;
    #  x = ---- 9--8 --7- -6-- 5--4 --3- -2-- 1--0
    # x = (x | (x >> 2)) & 0x030C30C3
    # x = ---- --98 ---- 76-- --54 ---- 32-- --10
    # x = (x | (x >> 4)) & 0x0300F00F
    # x = ---- --98 ---- ---- 7654 ---- ---- 3210
    # x = (x | (x >> 8)) & 0x030000FF
    # x = ---- --98 ---- ---- ---- ---- 7654 3210
    # x = (x | (x >> 16)) & 0x000003FF
    # x = ---- ---- ---- ---- ---- --98 7654 3210



from . import gain
from . import acoustics
from . import freqop
from . import geometry
from . import io
from . import signals
from . import plot
from .boxBuilder import shoebox
from .mesh import CAD as meshCAD
import numpy as np
from re import findall

## General functions
def findInArray(array, value):
    """
    Find the index and associated value of the closest data point in the array.

    Parameters
    ----------
    array : numpy array
        Array to be scanned through.

    value : float or complex or other data type
        Value to find in the array.

    Returns
    -------
    ind_value : int
        Index of the closest data point to the given value.

    value_out : float or complex or other data type
        Associated value in the input array that is closest to the given value.

    Notes
    -----
    This function calculates the index and the associated value of the closest data point
    in the input 'array' to the specified 'value'. It finds the index corresponding to the
    data point with the smallest absolute difference from the given value and returns both
    the index and the associated value from the array.
    """

    ind_value = int((np.abs(array - value)).argmin())
    value_out = array[ind_value]
    return ind_value, value_out


def normMinMax(dataset, minValue, maxValue):
    """
    Linearly normalize a dataset between specified minimum and maximum boundaries.

    Parameters
    ----------
    dataset : numpy array
        Dataset to be normalized.
    minValue : float
        Minimum boundary for normalization.
    maxValue : float
        Maximum boundary for normalization.

    Returns
    -------
    normDataset : numpy array
        Normalized dataset with values between minValue and maxValue.

    Notes
    -----
    This function linearly normalizes the input 'dataset' between the specified 'minValue'
    and 'maxValue' boundaries. It rescales the values of the dataset so that they span the
    range from 'minValue' to 'maxValue', while maintaining the relative differences between
    data points. The resulting 'normDataset' array contains the normalized dataset.
    """
    normDataset = (maxValue-minValue) * (dataset-np.min(dataset))/(np.max(dataset)-np.min(dataset))+minValue
    return normDataset

def wrap(phase):
    return np.arctan2(np.sin(phase), np.cos(phase))


def parallel(*elements):
    Y = 0
    for i in range(len(elements)):
        Y += 1/elements[i]
    Z = 1/Y
    return Z

def extract_numbers_to_list(input_string):
    # Use regular expression to find all numbers in the input string
    numbers = findall(r'\d+', input_string)

    # Convert the matched strings to actual integers and store them in a list
    numbers_list = [int(number) for number in numbers]

    return numbers_list

def slice_array_into_parts(arr, num_parts):
    if len(arr) % num_parts != 0:
        raise ValueError("Array length is not divisible by the number of parts")

    part_size = len(arr) // num_parts
    sliced_parts = [arr[i:i + part_size] for i in range(0, len(arr), part_size)]
    return sliced_parts


"""Definition of scalar function spaces."""

from bempp_cl.helpers import timeit as _timeit

import numpy as _np
import numba as _numba
from bempp_cl.api import log


@_timeit
def p0_discontinuous_function_space(
    grid,
    support_elements=None,
    segments=None,
    swapped_normals=None,
    include_boundary_dofs=None,
    truncate_at_segment_edge=None,
):
    """Define a space of piecewise constant functions."""
    from .space import SpaceBuilder, _process_segments

    if include_boundary_dofs is not None:
        log("Setting include_boundary_dofs has no effect on this space type.", "warning")
    if truncate_at_segment_edge is not None:
        log(
            "Setting truncate_at_segment_edge has no effect on this space type.",
            "warning",
        )

    support, normal_multipliers = _process_segments(grid, support_elements, segments, swapped_normals)

    elements_in_support = _np.flatnonzero(support)
    support_size = len(elements_in_support)

    local2global = _np.zeros((grid.number_of_elements, 1), dtype="uint32")
    local2global[support] = _np.expand_dims(_np.arange(support_size, dtype="uint32"), 1)

    local_multipliers = _np.zeros((grid.number_of_elements, 1), dtype="float64")
    local_multipliers[support] = 1

    collocation_points = _np.array([[1.0 / 3], [1.0 / 3]])

    return (
        SpaceBuilder(grid)
        .set_codomain_dimension(1)
        .set_support(support)
        .set_normal_multipliers(normal_multipliers)
        .set_order(0)
        .set_is_localised(True)
        .set_shapeset("p0_discontinuous")
        .set_identifier("p0_discontinuous")
        .set_local2global(local2global)
        .set_local_multipliers(local_multipliers)
        .set_collocation_points(collocation_points)
        .set_barycentric_representation(p0_barycentric_discontinuous_function_space)
        .set_numba_surface_gradient(_numba_p0_surface_gradient)
        .build()
    )


@_timeit
def p0_barycentric_discontinuous_function_space(coarse_space):
    """Define a space of piecewise constant functions over a barycentric grid."""
    from .space import SpaceBuilder
    from scipy.sparse import coo_matrix

    number_of_support_elements = coarse_space.number_of_support_elements
    bary_grid_number_of_elements = 6 * coarse_space.grid.number_of_elements

    bary_support_elements = 6 * _np.repeat(coarse_space.support_elements, 6) + _np.tile(
        _np.arange(6), number_of_support_elements
    )

    bary_support_size = len(bary_support_elements)

    support = _np.zeros(6 * coarse_space.grid.number_of_elements, dtype=_np.bool_)
    support[bary_support_elements] = True

    normal_multipliers = _np.repeat(coarse_space.normal_multipliers, 6)

    coarse_dofs = _np.repeat(_np.arange(number_of_support_elements, dtype=_np.uint32), 6)
    bary_dofs = _np.arange(6 * number_of_support_elements, dtype=_np.uint32)
    values = _np.ones(6 * number_of_support_elements, dtype=_np.float64)

    local2global = _np.zeros((bary_grid_number_of_elements, 1), dtype="uint32")
    local_multipliers = _np.zeros((bary_grid_number_of_elements, 1), dtype="uint32")

    local2global[support] = _np.arange(bary_support_size).reshape(bary_support_size, 1)

    local_multipliers[support] = 1

    transform = coo_matrix(
        (values, (bary_dofs, coarse_dofs)),
        shape=(bary_support_size, number_of_support_elements),
        dtype=_np.float64,
    ).tocsr()

    dof_transformation = transform @ coarse_space.map_to_localised_space

    collocation_points = _np.array([[1.0 / 3], [1.0 / 3]])

    return (
        SpaceBuilder(coarse_space.grid.barycentric_refinement)
        .set_codomain_dimension(1)
        .set_support(support)
        .set_normal_multipliers(normal_multipliers)
        .set_order(0)
        .set_is_localised(True)
        .set_is_barycentric(True)
        .set_shapeset("p0_discontinuous")
        .set_identifier("p0_discontinuous")
        .set_local2global(local2global)
        .set_local_multipliers(local_multipliers)
        .set_collocation_points(collocation_points)
        .set_dof_transformation(dof_transformation)
        .set_numba_surface_gradient(_numba_p0_surface_gradient)
        .build()
    )


@_timeit
def p1_discontinuous_function_space(
    grid,
    support_elements=None,
    segments=None,
    swapped_normals=None,
    include_boundary_dofs=None,
    truncate_at_segment_edge=None,
):
    """Define a discontinuous space of piecewise linear functions."""
    from .space import SpaceBuilder, _process_segments

    if include_boundary_dofs is not None:
        log("Setting include_boundary_dofs has no effect on this space type.", "warning")
    if truncate_at_segment_edge is not None:
        log(
            "Setting truncate_at_segment_edge has no effect on this space type.",
            "warning",
        )

    support, normal_multipliers = _process_segments(grid, support_elements, segments, swapped_normals)

    elements_in_support = _np.flatnonzero(support)
    support_size = len(elements_in_support)

    local2global = _np.zeros((grid.number_of_elements, 3), dtype="uint32")
    local2global[support] = _np.arange(3 * support_size).reshape(support_size, 3)

    local_multipliers = _np.zeros((grid.number_of_elements, 3), dtype="float64")
    local_multipliers[support] = 1

    return (
        SpaceBuilder(grid)
        .set_codomain_dimension(1)
        .set_support(support)
        .set_normal_multipliers(normal_multipliers)
        .set_order(1)
        .set_is_localised(True)
        .set_shapeset("p1_discontinuous")
        .set_identifier("p1_discontinuous")
        .set_local2global(local2global)
        .set_local_multipliers(local_multipliers)
        .set_numba_surface_gradient(_numba_p1_surface_gradient)
        .build()
    )


@_timeit
def p1_continuous_function_space(
    grid,
    support_elements=None,
    segments=None,
    swapped_normals=None,
    include_boundary_dofs=False,
    truncate_at_segment_edge=True,
):
    """Define a space of continuous piecewise linear functions."""
    from .space import SpaceBuilder, _process_segments

    support, normal_multipliers = _process_segments(grid, support_elements, segments, swapped_normals)

    # Create list of vertex neighbors. Needed for dofmap computation

    # vertex_neighbors = [[] for _ in range(grid.number_of_vertices)]
    # for index in range(grid.number_of_elements):
    # for vertex in grid.elements[:, index]:
    # vertex_neighbors[vertex].append(index)
    # vertex_neighbors1, index_ptr1 = serialise_list_of_lists(vertex_neighbors)

    vertex_neighbors, index_ptr = grid.vertex_neighbors

    local2global, local_multipliers, support = _compute_p1_dof_map(
        grid.data(),
        support,
        include_boundary_dofs,
        truncate_at_segment_edge,
        vertex_neighbors,
        index_ptr,
    )

    return (
        SpaceBuilder(grid)
        .set_codomain_dimension(1)
        .set_support(support)
        .set_normal_multipliers(normal_multipliers)
        .set_order(1)
        .set_is_localised(False)
        .set_shapeset("p1_discontinuous")
        .set_identifier("p1_continuous")
        .set_local2global(local2global)
        .set_local_multipliers(local_multipliers)
        .set_barycentric_representation(p1_barycentric_continuous_function_space)
        .set_numba_surface_gradient(_numba_p1_surface_gradient)
        .build()
    )


@_timeit
def p1_barycentric_continuous_function_space(coarse_space):
    """Define a space of piecewise constant functions over a barycentric grid."""
    from .space import SpaceBuilder
    from scipy.sparse import coo_matrix

    number_of_support_elements = coarse_space.number_of_support_elements
    bary_grid_number_of_elements = 6 * coarse_space.grid.number_of_elements

    bary_support_elements = 6 * _np.repeat(coarse_space.support_elements, 6) + _np.tile(
        _np.arange(6), number_of_support_elements
    )

    bary_support_size = len(bary_support_elements)

    support = _np.zeros(6 * coarse_space.grid.number_of_elements, dtype=_np.bool_)
    support[bary_support_elements] = True

    normal_multipliers = _np.repeat(coarse_space.normal_multipliers, 6)

    coeffs = [
        _np.array(
            [
                [1.0, 1 / 3, 1 / 2],
                [1.0, 1 / 2, 1 / 3],
                [0.0, 1 / 3, 1 / 2],
                [0.0, 0.0, 1 / 3],
                [0.0, 1 / 3, 0.0],
                [0.0, 1 / 2, 1 / 3],
            ]
        ),
        _np.array(
            [
                [0.0, 1 / 3, 0.0],
                [0.0, 1 / 2, 1 / 3],
                [1.0, 1 / 3, 1 / 2],
                [1.0, 1 / 2, 1 / 3],
                [0.0, 1 / 3, 1 / 2],
                [0.0, 0.0, 1 / 3],
            ]
        ),
        _np.array(
            [
                [0.0, 1 / 3, 1 / 2],
                [0.0, 0.0, 1 / 3],
                [0.0, 1 / 3, 0.0],
                [0.0, 1 / 2, 1 / 3],
                [1.0, 1 / 3, 1 / 2],
                [1.0, 1 / 2, 1 / 3],
            ]
        ),
    ]

    coarse_dofs, bary_dofs, values = generate_p1_map(coarse_space.grid.data(), coarse_space.support_elements, coeffs)

    local2global = _np.zeros((bary_grid_number_of_elements, 3), dtype="uint32")
    local_multipliers = _np.zeros((bary_grid_number_of_elements, 3), dtype="uint32")

    local2global[support] = _np.arange(3 * bary_support_size).reshape(bary_support_size, 3)

    local_multipliers[support] = 1

    transform = coo_matrix(
        (values, (bary_dofs, coarse_dofs)),
        shape=(3 * bary_support_size, 3 * number_of_support_elements),
        dtype=_np.float64,
    ).tocsr()

    dof_transformation = transform @ coarse_space.map_to_localised_space

    collocation_points = _np.array([[1.0 / 3], [1.0 / 3]])

    return (
        SpaceBuilder(coarse_space.grid.barycentric_refinement)
        .set_codomain_dimension(1)
        .set_support(support)
        .set_normal_multipliers(normal_multipliers)
        .set_order(0)
        .set_is_localised(True)
        .set_is_barycentric(True)
        .set_shapeset("p1_discontinuous")
        .set_identifier("p1_continuous")
        .set_local2global(local2global)
        .set_local_multipliers(local_multipliers)
        .set_collocation_points(collocation_points)
        .set_dof_transformation(dof_transformation)
        .set_numba_surface_gradient(_numba_p0_surface_gradient)
        .build()
    )


@_numba.njit(cache=True)
def generate_p1_map(grid_data, support_elements, coeffs):
    """Actually generate the sparse matrix data."""
    number_of_elements = len(support_elements)

    coarse_dofs = _np.empty(3 * 18 * number_of_elements, dtype=_np.uint32)
    bary_dofs = _np.empty(3 * 18 * number_of_elements, dtype=_np.uint32)
    values = _np.empty(3 * 18 * number_of_elements, dtype=_np.float64)

    # Iterate through the global dofs and fill up the
    # corresponding coefficients.

    count = 0

    for index, elem_index in enumerate(support_elements):
        # Assign the dofs for the six barycentric elements

        bary_elements = _np.arange(6) + 6 * index
        for local_dof in range(3):
            coarse_dof = 3 * index + local_dof
            bary_coeffs = coeffs[local_dof]
            coarse_dofs[count : count + 18] = coarse_dof
            bary_dofs[count : count + 18] = _np.arange(3 * bary_elements[0], 3 * bary_elements[0] + 18)
            values[count : count + 18] = bary_coeffs.ravel()
            count += 18
    return coarse_dofs, bary_dofs, values


@_timeit
@_numba.njit(cache=True)
def _compute_p1_dof_map(
    grid_data,
    support,
    include_boundary_dofs,
    truncate_at_segment_edge,
    vertex_neighbors,
    index_ptr,
):
    """Compute the local2global and local_multipliers maps for P1 space."""

    def find_index(array, value):
        """Return first position of value in array."""
        for index, val in enumerate(array):
            if val == value:
                return index
        return -1

    elements_in_support = []
    for index, val in enumerate(support):
        if val:
            elements_in_support.append(index)

    number_of_elements = grid_data.elements.shape[1]
    number_of_vertices = grid_data.vertices.shape[1]
    local2global = -_np.ones((number_of_elements, 3), dtype=_np.int32)

    vertex_is_dof = _np.zeros(number_of_vertices, dtype=_np.bool_)
    extended_support = []

    for element_index in elements_in_support:
        for local_index in range(3):
            vertex = grid_data.elements[local_index, element_index]
            neighbors = vertex_neighbors[index_ptr[vertex] : index_ptr[vertex + 1]]
            non_support_neighbors = [n for n in neighbors if not support[n]]
            node_is_interior = len(non_support_neighbors) == 0 and not grid_data.vertex_on_boundary[vertex]
            if include_boundary_dofs or node_is_interior:
                # Just add dof
                local2global[element_index, local_index] = vertex
                vertex_is_dof[vertex] = True
            if len(non_support_neighbors) > 0 and not truncate_at_segment_edge and include_boundary_dofs:
                for en in non_support_neighbors:
                    extended_support.append(en)
                    other_local_index = find_index(grid_data.elements[:, en], vertex)
                    local2global[en, other_local_index] = vertex
                    # vertex_is_dof was already set to True in previous if.

    # We have now all the vertices that have dofs attached and local2global
    # has the vertex index if it is used and -1 otherwise.

    # Now need to convert vertex indices into actual dof indices. For subgrids
    # these two are not identical.

    support_final = _np.zeros(number_of_elements, dtype=_np.bool_)
    local2global_final = _np.zeros((number_of_elements, 3), dtype=_np.uint32)
    local_multipliers = _np.zeros((number_of_elements, 3), dtype=_np.float64)

    dofs = -_np.ones(number_of_vertices)
    used_dofs = _np.flatnonzero(vertex_is_dof)
    global_dof_count = len(used_dofs)
    dofs[used_dofs] = _np.arange(global_dof_count)

    # Iterate through all support elements and replace vertex indices by
    # dof indices

    elements_in_support.extend(set(extended_support))

    for element_index in elements_in_support:
        for local_index in range(3):
            vertex_index = local2global[element_index, local_index]
            if vertex_index == -1:
                continue
            mapped_dof = dofs[vertex_index]
            support_final[element_index] = True
            local2global_final[element_index, local_index] = mapped_dof
            local_multipliers[element_index, local_index] = 1
        # If not every local index was used in a grid we need to
        # map the non-used local dofs to some global dof. Use the
        # one with maximal index in element. The corresponding
        # multipliers are set to zero so that these artificial dofs
        # do not influence computations.
        if support_final[element_index]:
            max_dof = _np.max(local2global_final[element_index])
            for local_index in range(3):
                if local2global[element_index, local_index] == -1:
                    local2global_final[element_index, local_index] = max_dof

    return local2global_final, local_multipliers, support_final


@_numba.njit
def _numba_p0_surface_gradient(
    element_index,
    shapeset_gradient,
    local_coordinates,
    grid_data,
    local_multipliers,
    normal_multipliers,
):
    """Evaluate the surface gradient."""
    return _np.zeros((1, 3, 1, local_coordinates.shape[1]), dtype=_np.float64)


@_numba.njit
def _numba_p1_surface_gradient(
    element_index,
    shapeset_gradient,
    local_coordinates,
    grid_data,
    local_multipliers,
    normal_multipliers,
):
    """Evaluate the surface gradient."""
    reference_values = shapeset_gradient(local_coordinates)
    result = _np.empty((1, 3, 3, local_coordinates.shape[1]), dtype=_np.float64)
    for index in range(3):
        result[0, :, index, :] = grid_data.jac_inv_trans[element_index].dot(reference_values[0, :, index, :])
    return result



"""
Collection of electric circuit blocks

"""

from electroacPy.circuitSolver.components.electric import inductance, capacitance
import random, string
from numpy import pi, sqrt

def randomblock_id(length):
   letters = string.ascii_lowercase
   return ''.join(random.choice(letters) for i in range(length))


class lowpass_butter:
    def __init__(self, A, B, order, fc, Re):
        """
        Create a lowpass-filter block. From Leo Beranek and Tim Mellow - 
        Acoustics Sound Fields, Transducers and Vibration.
        
        Parameters
        ----------
        np : int or str
            positive connection.
        nm : int or str
            negative connection.
        order : int
            order of the low-pass filter.
        fc : float
            cut-off frequency - up to order 6.
        Re : float
            Resistance of connected loudspeaker 
            (taken into account for component selection).

        Returns
        -------
        Values of capacitance and inductance.

        """
        self.A = A
        self.B = B
        self.order = order
        self.fc = fc
        self.Re = Re
        
        om0 = 2 * pi *self.fc
        L = Re / om0
        C = 1 / om0 / Re
        rnd_id = randomblock_id(3)
        np = str(A)
        nm = str(B)
        
        self.network = None

        if self.order == 1:
            self.network = {"L1": inductance(np, nm, L)}
            self.L = L
        
        elif self.order == 2:
            L1 = 2 * L
            C1 = 0.5 * C
            self.network = {"L1": inductance(np, nm, 2*L),
                            "C1": capacitance(nm, 0, 0.5*C)}
            self.L = L1
            self.C = C1 
        
        elif self.order == 3:
            L1 = 3/2 * L
            L2 = 1/2 * L
            C1 = 4/3 * C
            self.network = {"L1": inductance(np, np+'_lp3_'+rnd_id, 3/2*L),
                            "C1": capacitance(np+'_lp3_'+rnd_id, 0, 4/3*C),
                            "L2": inductance(np+'_lp3_'+rnd_id, nm, 1/2*L)}
            self.L = [L1, L2]
            self.C = [C1]
        
        elif self.order == 4:
            L1 = 4*sqrt(2)/3*L
            L2 = 2*sqrt(2)/3*L
            C1 = 9/4/sqrt(2)*C
            C2 = 1/2/sqrt(2)*C
            self.network = {"L1": inductance(np, np+'_lp3_'+rnd_id, L1),
                            "C1": capacitance(np+'_lp3_'+rnd_id, 0, C1),
                            "L2": inductance(np+'_lp3_'+rnd_id, nm, L2),
                            "C2": capacitance(nm, 0, C2)}
            self.L = [L1, L2]
            self.C = [C1, C2]
        
        elif self.order == 5:
            L1 = 5/(sqrt(5)+1)*L
            L2 = 2 * sqrt(5) / (sqrt(5) + 1) * L
            L3 = 1 / (sqrt(5) + 1) * L
            C1 = (2/sqrt(5) + 1) * 2 / sqrt(5) * C
            C2 = 2 / sqrt(5) * C
            self.network = {"L1": inductance(np, np+'_1lp3_'+rnd_id, L1),
                            "C1": capacitance(np+'_1lp3_'+rnd_id, 0, C1),
                            "L2": inductance(np+'_1lp3_'+rnd_id, 
                                              np+'_2lp3_'+rnd_id, L2),
                            "C2": capacitance(np+'_2lp3_'+rnd_id, 0, C2),
                            "L3": inductance(np+'_2lp3_'+rnd_id, nm, L3)}
            self.L = [L1, L2, L3]
            self.C = [C1, C2]
        
        elif self.order == 6:
            L1 = 9/5*L
            L2 = 81/55 * L
            L3 = 8/11 * L
            C1 = 50/27 * C
            C2 = 121/108 * C
            C3 = 1/4 * C
            self.network = {"L1": inductance(np, np+'_1lp3_'+rnd_id, L1),
                            "C1": capacitance(np+'_1lp3_'+rnd_id, 0, C1),
                            "L2": inductance(np+'_1lp3_'+rnd_id, 
                                              np+'_2lp3_'+rnd_id, L2),
                            "C2": capacitance(np+'_2lp3_'+rnd_id, 0, C2),
                            "L3": inductance(np+'_2lp3_'+rnd_id, nm, L3),
                            "C3": capacitance(nm, 0, C3)}
            self.L = [L1, L2, L3]
            self.C = [C1, C2, C3]
        else:
            raise Exception("Order cannot be higher than 6")

class highpass_butter:
    def __init__(self, A, B, order, fc, Re):
        """
        Create a highpass-filter block. From Beranek and Tim Mellow - 
        Acoustics Sound Fields, Transducers and Vibration.

        Parameters
        ----------
        np : int or str
            positive connection.
        nm : int or str
            negative connection.
        order : int
            order of the low-pass filter.
        fc : float
            cut-off frequency - up to order 6.
        Re : float
            Resistance of connected loudspeaker 
            (taken into account for component selection).

        Returns
        -------
        Values of capacitance and inductance.

        """
        self.A = A
        self.B = B
        self.order = order
        self.fc = fc
        self.Re = Re
        
        om0 = 2 * pi *self.fc
        L = Re / om0
        C = 1 / om0 / Re
        rnd_id = randomblock_id(3)
        np = str(A)
        nm = str(B)

        if self.order == 1:
            self.network = {"C1": capacitance(np, nm, C)}
            self.C = C
        
        elif self.order == 2:
            L1 = 2 * L
            C1 = 0.5 * C
            self.network = {"C1": capacitance(np, nm, C1),
                            "L1": inductance(nm, 0, L1)}
            self.L = L1
            self.C = C1 
        
        elif self.order == 3:
            C1 = 2/3 * C
            C2 = 2 * C
            L1 = 3/4 * L
            self.network = {"C1": capacitance(np, np+'_lp3_'+rnd_id, C1),
                            "L1": inductance(np+'_lp3_'+rnd_id, 0, L1),
                            "C2": capacitance(np+'_lp3_'+rnd_id, nm, C2)}
            self.L = [L1]
            self.C = [C1, C2]
        
        elif self.order == 4:
            C1 = 3/4/sqrt(2)*C
            C2 = 3/2/sqrt(2)*C
            L1 = 4*sqrt(2)/9*L
            L2 = 2*sqrt(2)*L
            self.network = {"C1": capacitance(np, np+'_lp3_'+rnd_id, C1),
                            "L1": inductance(np+'_lp3_'+rnd_id, 0, L1),
                            "C2": capacitance(np+'_lp3_'+rnd_id, nm, C2),
                            "L2": inductance(nm, 0, L2)}
            self.L = [L1, L2]
            self.C = [C1, C2]
        
        elif self.order == 5:
            C1 = (sqrt(5)+1)/5 * C
            C2 = (sqrt(5)+1)/2/sqrt(5) * C
            C3 = (sqrt(5)+1) * C
            L1 = 1/(2/sqrt(5)+1) * sqrt(5)/2 * L
            L2 = sqrt(5)/2 * L
            self.network = {"C1": capacitance(np, np+'_1lp3_'+rnd_id, C1),
                            "L1": inductance(np+'_1lp3_'+rnd_id, 0, L1),
                            "C2": capacitance(np+'_1lp3_'+rnd_id, 
                                              np+'_2lp3_'+rnd_id, C2),
                            "L2": inductance(np+'_2lp3_'+rnd_id, 0, L2),
                            "C3": capacitance(np+'_2lp3_'+rnd_id, nm, C3)}
            self.L = [L1, L2]
            self.C = [C1, C2, C3]
        
        elif self.order == 6:
            C1 = 9/5 * C
            C2 = 55/81 * C
            C3 = 11/8 * C
            L1 = 27/50*L
            L2 = 108/121 * L
            L3 = 4 * L
            self.network = {"C1": capacitance(np, np+'_1lp3_'+rnd_id, C1),
                            "L1": inductance(np+'_1lp3_'+rnd_id, 0, L1),
                            "C2": capacitance(np+'_1lp3_'+rnd_id, 
                                              np+'_2lp3_'+rnd_id, C2),
                            "L2": inductance(np+'_2lp3_'+rnd_id, 0, L2),
                            "C3": capacitance(np+'_2lp3_'+rnd_id, nm, C3),
                            "L3": inductance(nm, 0, L3)}
            self.L = [L1, L2, L3]
            self.C = [C1, C2, C3]
        else:
            raise Exception("Order cannot be higher than 6.")


"""Dense assembly."""

import numpy as _np
from bempp_cl.api.assembly import assembler as _assembler


class DenseAssembler(_assembler.AssemblerBase):
    """Implementation of a dense assembler for integral operators."""

    # pylint: disable=useless-super-delegation
    def __init__(self, domain, dual_to_range, parameters=None):
        """Create a dense assembler instance."""
        super().__init__(domain, dual_to_range, parameters)

    def assemble(self, operator_descriptor, device_interface, precision, *args, **kwargs):
        """Dense assembly of the integral operator."""
        from bempp_cl.api.assembly.discrete_boundary_operator import (
            DenseDiscreteBoundaryOperator,
        )
        from bempp_cl.api.utils.helpers import promote_to_double_precision

        if self.domain.requires_dof_transformation or self.dual_to_range.requires_dof_transformation:
            raise ValueError("Spaces that require dof transformations not supported for dense assembly.")

        # (
        # numba_assembly_function_regular,
        # numba_kernel_function_regular,
        # ) = select_numba_kernels(operator_descriptor, mode="regular")

        # (
        # numba_assembly_function_singular,
        # numba_kernel_function_singular,
        # ) = select_numba_kernels(operator_descriptor, mode="singular")

        mat = assemble_dense(
            self.domain,
            self.dual_to_range,
            self.parameters,
            operator_descriptor,
            device_interface,
        )

        if self.parameters.assembly.always_promote_to_double:
            mat = promote_to_double_precision(mat)

        return DenseDiscreteBoundaryOperator(mat)


def assemble_dense(domain, dual_to_range, parameters, operator_descriptor, device_interface):
    """Assembles the operator and returns a dense matrix."""
    import bempp_cl.api
    from bempp_cl.api.utils.helpers import get_type
    from bempp_cl.core.dispatcher import dense_assembler_dispatcher
    from bempp_cl.core.singular_assembler import assemble_singular_part

    precision = operator_descriptor.precision

    rows = dual_to_range.global_dof_count
    cols = domain.global_dof_count

    if operator_descriptor.is_complex:
        result_type = get_type(precision).complex
    else:
        result_type = get_type(precision).real

    result = _np.zeros((rows, cols), dtype=result_type)

    with bempp_cl.api.Timer(message=f"Regular assembler:{operator_descriptor.identifier}:{device_interface}"):
        dense_assembler_dispatcher(
            device_interface,
            operator_descriptor,
            domain,
            dual_to_range,
            parameters,
            result,
        )

    grids_identical = domain.grid == dual_to_range.grid

    if grids_identical:
        trial_local2global = domain.local2global.ravel()
        test_local2global = dual_to_range.local2global.ravel()
        trial_multipliers = domain.local_multipliers.ravel()
        test_multipliers = dual_to_range.local_multipliers.ravel()

        singular_rows, singular_cols, singular_values = assemble_singular_part(
            domain.localised_space,
            dual_to_range.localised_space,
            parameters,
            operator_descriptor,
            device_interface,
        )

        rows = test_local2global[singular_rows]
        cols = trial_local2global[singular_cols]
        values = singular_values * trial_multipliers[singular_cols] * test_multipliers[singular_rows]

        _np.add.at(result, (rows, cols), values)

    return result


# @_timeit
# def assemble_dense(
# domain,
# dual_to_range,
# parameters,
# operator_descriptor,
# numba_assembly_function_regular,
# numba_kernel_function_regular,
# numba_assembly_function_singular,
# numba_kernel_function_singular,
# ):
# """
# Really assemble the operator.
# Assembles the complete operator (near-field and far-field)
# Returns a dense matrix.
# """
# from bempp_cl.api.integration.triangle_gauss import rule as regular_rule
# from bempp_cl.api import log
# from bempp_cl.core.numba.singular_assembler import assemble_singular_part
# from bempp_cl.api.utils.helpers import get_type
# import bempp_cl.api

# order = parameters.quadrature.regular
# quad_points, quad_weights = regular_rule(order)

# test_indices, test_color_indexptr = dual_to_range.get_elements_by_color()
# trial_indices, trial_color_indexptr = domain.get_elements_by_color()
# number_of_test_colors = len(test_color_indexptr) - 1
# number_of_trial_colors = len(trial_color_indexptr) - 1

# rows = dual_to_range.global_dof_count
# cols = domain.global_dof_count

# nshape_test = dual_to_range.number_of_shape_functions
# nshape_trial = domain.number_of_shape_functions

# precision = operator_descriptor.precision

# data_type = get_type(precision).real
# if operator_descriptor.is_complex:
# result_type = get_type(precision).complex
# else:
# result_type = get_type(precision).real

# result = _np.zeros((rows, cols), dtype=result_type)

# grids_identical = domain.grid == dual_to_range.grid

# dense_assembler_dispatcher(
# device_interface,
# operator_descriptor,
# domain,
# dual_to_range,


# with bempp_cl.api.Timer(message="Start Numba dense assembler."):
# for test_color_index in range(number_of_test_colors):
# numba_assembly_function_regular(
# dual_to_range.grid.data(precision),
# domain.grid.data(precision),
# nshape_test,
# nshape_trial,
# test_indices[
# test_color_indexptr[test_color_index] : test_color_indexptr[
# 1 + test_color_index
# ]
# ],
# trial_indices,
# dual_to_range.local_multipliers.astype(data_type),
# domain.local_multipliers.astype(data_type),
# dual_to_range.local2global,
# domain.local2global,
# dual_to_range.normal_multipliers,
# domain.normal_multipliers,
# quad_points.astype(data_type),
# quad_weights.astype(data_type),
# numba_kernel_function_regular,
# _np.array(operator_descriptor.options, dtype=data_type),
# grids_identical,
# dual_to_range.shapeset.evaluate,
# domain.shapeset.evaluate,
# result,
# )

# if grids_identical:
# # Need to treat singular contribution

# trial_local2global = domain.local2global.ravel()
# test_local2global = dual_to_range.local2global.ravel()
# trial_multipliers = domain.local_multipliers.ravel()
# test_multipliers = dual_to_range.local_multipliers.ravel()

# singular_rows, singular_cols, singular_values = assemble_singular_part(
# domain.localised_space,
# dual_to_range.localised_space,
# parameters,
# operator_descriptor,
# numba_assembly_function_singular,
# numba_kernel_function_singular,
# )

# rows = test_local2global[singular_rows]
# cols = trial_local2global[singular_cols]
# values = (
# singular_values
# * trial_multipliers[singular_cols]
# * test_multipliers[singular_rows]
# )

# _np.add.at(result, (rows, cols), values)

# return result



"""Implementation of boundary operators."""


class BoundaryOperator(object):
    """A base class for boundary operators."""

    def __init__(self, domain, range_, dual_to_range, parameters):
        """Construct. Should only be called through derived class."""
        self._domain = domain
        self._range = range_
        self._dual_to_range = dual_to_range
        self._parameters = parameters
        self._cached = None
        self._range_map = None

    @property
    def domain(self):
        """Return the domain space of the operator."""
        return self._domain

    @property
    def range(self):
        """Return the range space of the operator."""
        return self._range

    @property
    def dual_to_range(self):
        """Return the dual to range space of the operator."""
        return self._dual_to_range

    @property
    def parameters(self):
        """Return the parameters associated with the operator."""
        return self._parameters

    def weak_form(self):
        """Return the weak form (assemble if necessary)."""
        if not self._cached:
            self._cached = self._assemble()

        return self._cached

    def strong_form(self):
        """Return a discrete operator that maps into the range space."""
        from bempp_cl.api.utils.helpers import get_inverse_mass_matrix

        if self._range_map is None:
            # This is the most frequent case and we cache the mass
            # matrix from the space object.
            self._range_map = get_inverse_mass_matrix(self.range, self.dual_to_range)

        return self._range_map * self.weak_form()

    def __mul__(self, other):
        """Return product with a scalar, grid function or other operator."""
        import numpy as np
        from bempp_cl.api import GridFunction

        if np.isscalar(other):
            return _ScaledBoundaryOperator(self, other)
        elif isinstance(other, BoundaryOperator):
            return _ProductBoundaryOperator(self, other)
        elif isinstance(other, GridFunction):
            if not self.domain.is_compatible(other.space):
                raise ValueError("Operator domain space does not match GridFunction space.")
            return GridFunction(
                self.range,
                projections=self.weak_form() * other.coefficients,
                dual_space=self.dual_to_range,
            )
        else:
            return NotImplemented

    def __matmul__(self, other):
        """Multiply by another operator."""
        return self.__mul__(other)

    def __rmul__(self, other):
        """Multiply by another operator."""
        import numpy as np

        if np.isscalar(other):
            return _ScaledBoundaryOperator(self, other)
        else:
            return NotImplemented

    def __add__(self, other):
        """Return sum of two boundary operators."""
        return _SumBoundaryOperator(self, other)

    def __neg__(self):
        """Negate an operator."""
        return self.__mul__(-1.0)

    def __sub__(self, other):
        """Subtract two operators."""
        return self.__add__(-other)


class BoundaryOperatorWithAssembler(BoundaryOperator):
    """Implements a boundary operator together with an assembler."""

    # pylint: disable=too-many-arguments
    def __init__(self, domain, range_, dual_to_range, assembler, operator_descriptor, transpose=False):
        """Initialize a boundary operator with assembler."""
        super().__init__(domain, range_, dual_to_range, assembler.parameters)

        self._assembler = assembler
        self._operator_descriptor = operator_descriptor
        self.transpose_ = transpose

    @property
    def assembler(self):
        """Return the assembler associated with this operator."""
        return self._assembler

    @property
    def descriptor(self):
        """Operator descriptor."""
        return self._operator_descriptor

    def _assemble(self):
        """Assemble the operator."""
        if self.transpose_:
            return self.assembler.assemble(self.descriptor).transpose()
        return self.assembler.assemble(self.descriptor)

    def _transpose(self, _range):
        return BoundaryOperatorWithAssembler(
            self._dual_to_range, self._domain, _range, self._assembler, self._operator_descriptor, True
        )


class _SumBoundaryOperator(BoundaryOperator):
    """Return the sum of two boundary operators."""

    def __init__(self, op1, op2):
        """Construct the sum of two boundary operators."""
        if (
            not op1.domain.is_compatible(op2.domain)
            or not op1.range.is_compatible(op2.range)
            or not op1.dual_to_range.is_compatible(op2.dual_to_range)
        ):
            raise ValueError("Spaces not compatible.")

        super(_SumBoundaryOperator, self).__init__(op1.domain, op1.range, op1.dual_to_range, None)

        self._op1 = op1
        self._op2 = op2

    def _assemble(self):
        """Implement the weak form."""
        return self._op1.weak_form() + self._op2.weak_form()


class _ScaledBoundaryOperator(BoundaryOperator):
    """Scale a boundary operator."""

    def __init__(self, op, alpha):
        """Construct a scaled boundary operator."""
        super(_ScaledBoundaryOperator, self).__init__(op.domain, op.range, op.dual_to_range, op.parameters)

        self._op = op
        self._alpha = alpha

    def _assemble(self):
        """Implement the weak form."""
        return self._op.weak_form() * self._alpha


class _ProductBoundaryOperator(BoundaryOperator):
    """Multiply two boundary operators."""

    def __init__(self, op1, op2):
        """Construct a product boundary operator."""
        if not op2.range.is_compatible(op1.domain):
            raise ValueError(
                "Range space of second operator must be compatible to " + "domain space of first operator."
            )

        super(_ProductBoundaryOperator, self).__init__(op2.domain, op1.range, op1.dual_to_range, None)

        self._op1 = op1
        self._op2 = op2

    def _assemble(self):
        """Implement the weak form."""
        return self._op1.weak_form() * self._op2.strong_form()


class ZeroBoundaryOperator(BoundaryOperator):
    """A boundary operator that represents a zero operator.

    Parameters
    ----------
    domain : bempp_cl.api.space.Space
        Domain space of the operator.
    range_ : bempp_cl.api.space.Space
        Range space of the operator.
    dual_to_range : bempp_cl.api.space.Space
        Dual space to the range space.

    """

    def __init__(self, domain, range_, dual_to_range):
        import bempp_cl.api

        super(ZeroBoundaryOperator, self).__init__(domain, range_, dual_to_range, bempp_cl.api.GLOBAL_PARAMETERS)

    def _assemble(self):
        from bempp_cl.api.assembly.discrete_boundary_operator import (
            ZeroDiscreteBoundaryOperator,
        )

        return ZeroDiscreteBoundaryOperator(self.dual_to_range.global_dof_count, self.domain.global_dof_count)

    def __iadd__(self, other):
        """Add."""
        if self.domain != other.domain or self.range != other.range or self.dual_to_range != other.dual_to_range:
            raise ValueError("Spaces not compatible.")

        return other

    def __isub__(self, other):
        """Subtract."""
        if self.domain != other.domain or self.range != other.range or self.dual_to_range != other.dual_to_range:
            raise ValueError("Spaces not compatible.")

        return -other


class MultiplicationOperator(BoundaryOperator):
    """
    Returns a weak multiplication operator with a grid function.

    Given a grid function g, this operator assembles a sparse
    matrix that implements weak multiplication with g in a given
    basis.
    """

    def __init__(
        self,
        grid_function,
        domain,
        range_,
        dual_to_range,
        parameters=None,
        mode="component",
    ):
        """
        Initialize the multiplication operator.

        This class initializes a multiplication operator mult from a given
        grid function g, such that the result h = mult @ f for a given
        grid function f is the result of projecting g * f onto the
        space `dual_to_range`.

        The number of components of the shapesets of grid_function.space,
        domain and dual_to_range must be compatible. For example, grid_function
        could be scalar, and space and dual_to_range vectorial with 3 components.


        Parameters
        ----------
        grid_function : GridFunction object
            The grid function object from which the multiplication operator is
            built. It needs to allow representation in primal (coordinate) form.
        domain : Space object
            The domain space of the operator.
        range_ : Space object
            The range space of the operator
        dual_to_range : Space object
            The dual space on which the result of the multiplication is projected.
        parameters : Parameters object
            The parameters associated with this operator.
        mode : string
            Either 'component' or 'inner'. If mode is 'component' (default) then
            the multiplication is componentwise at each quadrature point. Hence,
            if the shapesets of grid_function.space and domain.space both have
            dimension 3, then the result of the multiplication is also of dimension
            3, and the dual_to_range space needs to be compatible to it. If mode
            is 'inner', then inner product is taken of the values of grid_function
            at the quadrature point and the basis functions in space at the
            quadrature points.

        """
        from bempp_cl.api.utils.helpers import assign_parameters

        self._mode = mode
        self._grid_fun = grid_function

        # Check compatibility

        dim = grid_function.component_count
        dual_dim = dual_to_range.codomain_dimension
        dimensions_compatible = False
        if mode == "component":
            dimensions_compatible = dim == domain.codomain_dimension and dim == dual_dim
        elif mode == "inner":
            dimensions_compatible = dim == domain.codomain_dimension and dual_dim == 1
        else:
            raise ValueError("Unknown value for 'mode'. Allowed: 'component', 'inner'")
        if not dimensions_compatible:
            raise ValueError("Incompatible codomain dimensions.")

        super().__init__(domain, range_, dual_to_range, assign_parameters(parameters))

    def _assemble(self):
        """Assemble the operator."""
        from bempp_cl.api.space.space import return_compatible_representation
        from bempp_cl.api.assembly.discrete_boundary_operator import (
            SparseDiscreteBoundaryOperator,
        )
        from bempp_cl.api.integration.triangle_gauss import rule
        from scipy.sparse import coo_matrix

        import numpy as _np

        points, weights = rule(self._parameters.quadrature.regular)

        comp_trial, comp_test, comp_fun = return_compatible_representation(
            self.domain, self.dual_to_range, self._grid_fun.space
        )

        grid = comp_trial.grid

        if self._mode == "component":
            op = _np.multiply
        elif self._mode == "inner":
            op = lambda x, y: _np.sum(x * y.reshape[:, _np.newaxis, :], axis=0, keepdims=True)

        elements = (
            set(comp_test.support_elements)
            .intersection(set(comp_trial.support_elements))
            .intersection(set(comp_fun.support_elements))
        )

        elements = _np.flatnonzero(comp_trial.support * comp_test.support * comp_fun.support)

        number_of_elements = len(elements)
        nshape_trial = comp_trial.shapeset.number_of_shape_functions
        nshape_test = comp_test.shapeset.number_of_shape_functions
        nshape = nshape_trial * nshape_test

        if _np.iscomplexobj(self._grid_fun.coefficients):
            dtype = "complex128"
        else:
            dtype = "float64"

        data = _np.zeros(number_of_elements * nshape_trial * nshape_test, dtype=dtype)

        for index, elem_index in enumerate(elements):
            scale_vals = self._grid_fun.evaluate(elem_index, points) * weights * grid.integration_elements[index]
            domain_vals = comp_trial.evaluate(elem_index, points)
            trial_vals = op(domain_vals, scale_vals)
            test_vals = _np.conj(comp_test.evaluate(elem_index, points))
            res = _np.tensordot(test_vals, trial_vals, axes=([0, 2], [0, 2]))
            data[nshape * index : nshape * (1 + index)] = res.ravel()

        irange = _np.arange(nshape_test)
        jrange = _np.arange(nshape_trial)

        rows = _np.tile(_np.repeat(irange, nshape_trial), number_of_elements) + _np.repeat(
            elements * nshape_test, nshape
        )

        cols = _np.tile(_np.tile(jrange, nshape_test), number_of_elements) + _np.repeat(elements * nshape_trial, nshape)

        new_rows = comp_test.local2global.ravel()[rows]
        new_cols = comp_trial.local2global.ravel()[cols]

        nrows = comp_test.dof_transformation.shape[0]
        ncols = comp_trial.dof_transformation.shape[0]

        mat = coo_matrix((data, (new_rows, new_cols)), shape=(nrows, ncols)).tocsr()

        if comp_trial.requires_dof_transformation:
            mat = mat @ self.domain.dof_transformation

        if comp_test.requires_dof_transformation:
            mat = self.dual_to_range.dof_transformation.T @ mat

        return SparseDiscreteBoundaryOperator(mat)



"""FMM helper functions."""

import numpy as _np
import numba as _numba

M_INV_4PI = 1.0 / (4 * _np.pi)


@_numba.jit(nopython=True, parallel=False, error_model="numpy", fastmath=True, boundscheck=False)
def laplace_kernel(target_points, source_points, kernel_parameters, dtype, result_type):
    """Evaluate the Laplace kernel."""
    ntargets = target_points.shape[1]
    nsources = source_points.shape[1]
    m_inv_4pi = dtype.type(M_INV_4PI)

    interactions = _np.empty(4 * ntargets * nsources, dtype=result_type)
    diff = _np.zeros((3, nsources), dtype=dtype)
    for target_point_index in range(ntargets):
        dist = _np.zeros(nsources, dtype=dtype)
        for i in range(3):
            for j in range(nsources):
                diff[i, j] = target_points[i, target_point_index] - source_points[i, j]
                dist[j] += diff[i, j] * diff[i, j]
        for j in range(nsources):
            dist[j] = _np.sqrt(dist[j])
        for j in range(nsources):
            interactions[target_point_index * 4 * nsources + 4 * j] = m_inv_4pi / dist[j]
        for i in range(3):
            for j in range(nsources):
                interactions[target_point_index * 4 * nsources + 4 * j + 1 + i] = (
                    -diff[i, j] * m_inv_4pi / (dist[j] * dist[j] * dist[j])
                )
        # Now fix zero distance case
        for j in range(nsources):
            if dist[j] == 0:
                for i in range(4):
                    interactions[target_point_index * 4 * nsources + 4 * j + i] = 0

    return interactions


@_numba.jit(nopython=True, parallel=False, error_model="numpy", fastmath=True, boundscheck=False)
def modified_helmholtz_kernel(target_points, source_points, kernel_parameters, dtype, result_type):
    """Evaluate the modified Helmholtz kernel."""
    ntargets = target_points.shape[1]
    nsources = source_points.shape[1]
    m_inv_4pi = dtype.type(M_INV_4PI)

    interactions = _np.empty(4 * ntargets * nsources, dtype=result_type)
    diff = _np.zeros((3, nsources), dtype=dtype)
    for target_point_index in range(ntargets):
        dist = _np.zeros(nsources, dtype=dtype)
        for i in range(3):
            for j in range(nsources):
                diff[i, j] = target_points[i, target_point_index] - source_points[i, j]
                dist[j] += diff[i, j] * diff[i, j]
        for j in range(nsources):
            dist[j] = _np.sqrt(dist[j])
        for j in range(nsources):
            interactions[target_point_index * 4 * nsources + 4 * j] = (
                m_inv_4pi * _np.exp(-kernel_parameters[0] * dist[j]) / dist[j]
            )
        for i in range(3):
            for j in range(nsources):
                interactions[target_point_index * 4 * nsources + 4 * j + 1 + i] = (
                    (-kernel_parameters[0] * dist[j] - 1)
                    * diff[i, j]
                    * _np.exp(-kernel_parameters[0] * dist[j])
                    * m_inv_4pi
                    / (dist[j] * dist[j] * dist[j])
                )
        # Now fix zero distance case
        for j in range(nsources):
            if dist[j] == 0:
                for i in range(4):
                    interactions[target_point_index * 4 * nsources + 4 * j + i] = 0

    return interactions


@_numba.jit(nopython=True, parallel=False, error_model="numpy", fastmath=True, boundscheck=False)
def helmholtz_kernel(target_points, source_points, kernel_parameters, dtype, result_type):
    """Evaluate the Laplace kernel."""
    ntargets = target_points.shape[1]
    nsources = source_points.shape[1]
    m_inv_4pi = dtype.type(M_INV_4PI)

    tmp_real = _np.empty(nsources, dtype=dtype)
    tmp_imag = _np.empty(nsources, dtype=dtype)
    interactions_real = _np.empty(4 * ntargets * nsources, dtype=dtype)
    interactions_imag = _np.empty(4 * ntargets * nsources, dtype=dtype)
    diff = _np.zeros((3, nsources), dtype=dtype)
    for target_point_index in range(ntargets):
        dist = _np.zeros(nsources, dtype=dtype)
        for i in range(3):
            for j in range(nsources):
                diff[i, j] = target_points[i, target_point_index] - source_points[i, j]
                dist[j] += diff[i, j] * diff[i, j]
        for j in range(nsources):
            dist[j] = _np.sqrt(dist[j])
        for j in range(nsources):
            tmp_real[j] = m_inv_4pi * _np.cos(kernel_parameters[0] * dist[j]) / dist[j]
            tmp_imag[j] = m_inv_4pi * _np.sin(kernel_parameters[0] * dist[j]) / dist[j]
        if kernel_parameters[1] != 0:
            for j in range(nsources):
                tmp_real[j] *= _np.exp(-kernel_parameters[1] * dist[j])
                tmp_imag[j] *= _np.exp(-kernel_parameters[1] * dist[j])
        for j in range(nsources):
            interactions_real[target_point_index * 4 * nsources + 4 * j] = tmp_real[j]
            interactions_imag[target_point_index * 4 * nsources + 4 * j] = tmp_imag[j]
        for i in range(3):
            for j in range(nsources):
                interactions_real[target_point_index * 4 * nsources + 4 * j + 1 + i] = (
                    (-(1 + dist[j] * kernel_parameters[1]) * tmp_real[j] - kernel_parameters[0] * dist[j] * tmp_imag[j])
                    / (dist[j] * dist[j])
                    * diff[i, j]
                )
                interactions_imag[target_point_index * 4 * nsources + 4 * j + 1 + i] = (
                    (-(1 + dist[j] * kernel_parameters[1]) * tmp_imag[j] + kernel_parameters[0] * dist[j] * tmp_real[j])
                    / (dist[j] * dist[j])
                    * diff[i, j]
                )

        # Now fix zero distance case
        for j in range(nsources):
            if dist[j] == 0:
                for i in range(4):
                    interactions_real[target_point_index * 4 * nsources + 4 * j + i] = 0
                    interactions_imag[target_point_index * 4 * nsources + 4 * j + i] = 0

    return interactions_real + 1j * interactions_imag


def get_local_interaction_operator(
    grid,
    local_points,
    kernel_function,
    kernel_parameters,
    precision,
    is_complex,
    device_interface=None,
):
    """Get the local interaction operator."""
    import bempp_cl.api
    from bempp_cl.api import GLOBAL_PARAMETERS
    from bempp_cl.api.utils.helpers import get_type
    from scipy.sparse import csr_matrix
    from scipy.sparse.linalg import aslinearoperator
    from scipy.sparse.linalg import LinearOperator

    npoints = local_points.shape[1]

    dtype = _np.dtype(get_type(precision).real)
    if is_complex:
        result_type = _np.dtype(get_type(precision).complex)
    else:
        result_type = dtype

    rows = 4 * npoints * grid.number_of_elements
    cols = npoints * grid.number_of_elements

    if kernel_function == "laplace":
        kernel = laplace_kernel
    elif kernel_function == "helmholtz":
        kernel = helmholtz_kernel
    elif kernel_function == "modified_helmholtz":
        kernel = modified_helmholtz_kernel

    if GLOBAL_PARAMETERS.fmm.near_field_representation == "sparse":
        data, indices, indexptr = get_local_interaction_matrix_impl(
            grid.data(precision),
            local_points.astype(dtype),
            kernel,
            _np.array(kernel_parameters, dtype=dtype),
            dtype,
            result_type,
        )
        return aslinearoperator(csr_matrix((data, indices, indexptr), shape=(rows, cols)))

    elif GLOBAL_PARAMETERS.fmm.near_field_representation == "evaluate":
        if device_interface is None:
            device_interface = bempp_cl.api.DEFAULT_DEVICE_INTERFACE
        if device_interface == "numba":
            evaluator = get_local_interaction_evaluator_numba(
                grid.data(precision),
                local_points.astype(dtype),
                kernel,
                _np.array(kernel_parameters, dtype=dtype),
                dtype,
                result_type,
            )
            return LinearOperator(shape=(rows, cols), matvec=evaluator, dtype=result_type)
        elif device_interface == "opencl":
            evaluator = get_local_interaction_evaluator_opencl(
                grid,
                local_points.astype(dtype),
                kernel_function,
                _np.array(kernel_parameters, dtype=dtype),
                dtype,
                result_type,
            )
            return LinearOperator(shape=(rows, cols), matvec=evaluator, dtype=result_type)
        else:
            raise ValueError("Device interface must be one of 'numba', 'opencl'.")
    else:
        raise ValueError("Unknown value for near_field_representation.")


def get_local_interaction_evaluator_numba(
    grid_data, local_points, kernel_function, kernel_parameters, dtype, result_type
):
    """Return an evaluator for the local interactions."""
    import bempp_cl.api

    def evaluator(coeffs):
        """Actually evaluate the near field correction."""
        with bempp_cl.api.Timer(message="Singular Corrections Evaluator."):
            return numba_evaluate_local_interactions(
                grid_data,
                coeffs,
                local_points,
                kernel_function,
                kernel_parameters,
                dtype,
                result_type,
            )

    return evaluator


@_numba.jit(nopython=True, parallel=True, error_model="numpy", fastmath=True, boundscheck=False)
def numba_evaluate_local_interactions(
    grid_data,
    coeffs,
    local_points,
    kernel_function,
    kernel_parameters,
    dtype,
    result_type,
):
    """Get the local interaction matrix on the grid."""
    nelements = grid_data.elements.shape[1]
    npoints = local_points.shape[1]
    neighbor_indices = grid_data.element_neighbor_indices
    neighbor_indexptr = grid_data.element_neighbor_indexptr

    result = _np.zeros(4 * npoints * nelements, dtype=result_type)
    global_points = _np.zeros((nelements, 3, npoints), dtype=dtype)

    for element_index in range(nelements):
        global_points[element_index, :, :] = grid_data.local2global(element_index, local_points)

    for target_element in _numba.prange(nelements):
        nneighbors = neighbor_indexptr[1 + target_element] - neighbor_indexptr[target_element]
        source_elements = _np.sort(
            neighbor_indices[neighbor_indexptr[target_element] : neighbor_indexptr[1 + target_element]]
        )

        local_source_points = _np.empty((3, npoints * nneighbors), dtype=dtype)
        for source_element_index in range(nneighbors):
            source_element = source_elements[source_element_index]
            local_source_points[:, npoints * source_element_index : npoints * (1 + source_element_index)] = (
                global_points[source_element, :, :]
            )
        local_target_points = global_points[target_element, :, :]
        interactions = kernel_function(
            local_target_points,
            local_source_points,
            kernel_parameters,
            dtype,
            result_type,
        )

        for target_point_index in range(npoints):
            for i in range(4):
                for source_element_index in range(nneighbors):
                    source_element = source_elements[source_element_index]
                    for source_point_index in range(npoints):
                        result[4 * npoints * target_element + 4 * target_point_index + i] += (
                            interactions[
                                4 * target_point_index * nneighbors * npoints
                                + 4 * source_element_index * npoints
                                + 4 * source_point_index
                                + i
                            ]
                            * coeffs[npoints * source_element + source_point_index]
                        )

    return result


@_numba.jit(nopython=True, parallel=True, error_model="numpy", fastmath=True, boundscheck=False)
def get_local_interaction_matrix_impl(grid_data, local_points, kernel_function, kernel_parameters, dtype, result_type):
    """Get the local interaction matrix on the grid."""
    nelements = grid_data.elements.shape[1]
    npoints = local_points.shape[1]
    neighbor_indices = grid_data.element_neighbor_indices
    neighbor_indexptr = grid_data.element_neighbor_indexptr

    data = _np.zeros(4 * npoints * npoints * len(neighbor_indices), dtype=result_type)
    indexptr = _np.zeros(4 * npoints * nelements + 1, dtype=_np.uint32)
    indices = _np.zeros(4 * npoints * npoints * len(neighbor_indices), dtype=_np.uint32)
    indexptr[-1] = 4 * npoints * npoints * len(neighbor_indices)

    global_points = _np.zeros((nelements, 3, npoints), dtype=dtype)

    for element_index in range(nelements):
        global_points[element_index, :, :] = grid_data.local2global(element_index, local_points)

    for target_element in _numba.prange(nelements):
        nneighbors = neighbor_indexptr[1 + target_element] - neighbor_indexptr[target_element]
        source_elements = _np.sort(
            neighbor_indices[neighbor_indexptr[target_element] : neighbor_indexptr[1 + target_element]]
        )

        local_source_points = _np.empty((3, npoints * nneighbors), dtype=dtype)
        for source_element_index in range(nneighbors):
            source_element = source_elements[source_element_index]
            local_source_points[:, npoints * source_element_index : npoints * (1 + source_element_index)] = (
                global_points[source_element, :, :]
            )
        local_target_points = global_points[target_element, :, :]
        interactions = kernel_function(
            local_target_points,
            local_source_points,
            kernel_parameters,
            dtype,
            result_type,
        )

        local_count = 4 * npoints * npoints * neighbor_indexptr[target_element]
        for target_point_index in range(npoints):
            for i in range(4):
                indexptr[4 * npoints * target_element + 4 * target_point_index + i] = local_count
                for source_element_index in range(nneighbors):
                    source_element = source_elements[source_element_index]
                    for source_point_index in range(npoints):
                        data[local_count] = interactions[
                            4 * target_point_index * nneighbors * npoints
                            + 4 * source_element_index * npoints
                            + 4 * source_point_index
                            + i
                        ]
                        indices[local_count] = npoints * source_element + source_point_index
                        local_count += 1

    return data, indices, indexptr


def map_space_to_points(space, local_points, weights, return_transpose=False):
    """Return mapper from grid coeffs to point evaluations."""
    from scipy.sparse import coo_matrix
    from scipy.sparse.linalg import aslinearoperator

    grid = space.grid
    number_of_local_points = local_points.shape[1]
    number_of_vertices = number_of_local_points * grid.number_of_elements

    data, global_indices, vertex_indices = map_space_to_points_impl(
        grid.data("double"),
        space.localised_space.local2global,
        space.localised_space.local_multipliers,
        space.localised_space.normal_multipliers,
        space.support_elements,
        space.numba_evaluate,
        space.shapeset.evaluate,
        local_points,
        weights,
        space.number_of_shape_functions,
    )

    if return_transpose:
        transform = coo_matrix(
            (data, (global_indices, vertex_indices)),
            shape=(space.localised_space.global_dof_count, number_of_vertices),
        )

        return aslinearoperator(space.map_to_localised_space.T) @ aslinearoperator(transform)
    else:
        transform = coo_matrix(
            (data, (vertex_indices, global_indices)),
            shape=(number_of_vertices, space.localised_space.global_dof_count),
        )
        return aslinearoperator(transform) @ aslinearoperator(space.map_to_localised_space)


@_numba.njit
def map_space_to_points_impl(
    grid_data,
    local2global,
    local_multipliers,
    normal_multipliers,
    support_elements,
    numba_evaluate,
    shape_fun,
    local_points,
    weights,
    number_of_shape_functions,
):
    """Numba accelerated computational parts for point map."""
    number_of_local_points = local_points.shape[1]
    number_of_support_elements = len(support_elements)

    nlocal = number_of_local_points * number_of_shape_functions

    data = _np.empty(nlocal * number_of_support_elements, dtype=_np.float64)
    global_indices = _np.empty(nlocal * number_of_support_elements, dtype=_np.int64)
    vertex_indices = _np.empty(nlocal * number_of_support_elements, dtype=_np.int64)

    for elem in support_elements:
        basis_values = (
            numba_evaluate(
                elem,
                shape_fun,
                local_points,
                grid_data,
                local_multipliers,
                normal_multipliers,
            )[0, :, :]
            * weights
            * grid_data.integration_elements[elem]
        )
        data[elem * nlocal : (1 + elem) * nlocal] = basis_values.ravel()
        for index in range(number_of_shape_functions):
            vertex_indices[
                elem * nlocal + index * number_of_local_points : elem * nlocal + (1 + index) * number_of_local_points
            ] = _np.arange(elem * number_of_local_points, (1 + elem) * number_of_local_points)
        global_indices[elem * nlocal : (1 + elem) * nlocal] = _np.repeat(local2global[elem, :], number_of_local_points)

    return (data, global_indices, vertex_indices)


@_numba.njit
def grid_to_points(grid_data, local_points):
    """
    Map a grid to an array of points.

    Returns a (N, 3) point array that stores the global vertices
    associated with the local points in each triangle.
    Points are stored in consecutive order for each element
    in the support_elements list. Hence, the returned array is of the form
    [ v_1^1, v_2^1, ..., v_M^1, v_1^2, v_2^2, ...], where
    v_i^j is the ith point in the jth element in
    the support_elements list.

    Parameters
    ----------
    grid_data : GridData
        A Bempp GridData object.
    local_points : np.ndarray
        (2, M) array of local coordinates.
    """
    number_of_elements = grid_data.elements.shape[1]
    number_of_points = local_points.shape[1]

    points = _np.empty((number_of_points * number_of_elements, 3), dtype=_np.float64)

    for elem in range(number_of_elements):
        points[number_of_points * elem : number_of_points * (1 + elem), :] = (
            _np.expand_dims(grid_data.vertices[:, grid_data.elements[0, elem]], 1)
            + grid_data.jacobians[elem].dot(local_points)
        ).T
    return points


def get_local_interaction_evaluator_opencl(grid, local_points, kernel_function, kernel_parameters, dtype, result_type):
    """Return an evaluator for the local interactions."""
    import pyopencl as _cl
    import bempp_cl.api
    from bempp_cl.core.opencl_kernels import get_kernel_from_name
    from bempp_cl.core.opencl_kernels import default_context, default_device

    if "laplace" in kernel_function:
        mode = "laplace"
    elif "modified_helmholtz" in kernel_function:
        mode = "modified_helmholtz"
    elif "helmholtz" in kernel_function:
        mode = "helmholtz"
    else:
        raise ValueError("Unknown value for kernel_function.")

    mf = _cl.mem_flags
    ctx = default_context()
    device = default_device()
    # vector_width = get_vector_width("double")
    npoints = local_points.shape[1]
    ncoeffs = npoints * grid.number_of_elements

    max_nneighbors = _np.max(_np.diff(grid.element_neighbors.indexptr))

    grid_buffer = _cl.Buffer(
        ctx,
        mf.READ_ONLY | mf.COPY_HOST_PTR,
        hostbuf=grid.as_array.astype(dtype),
    )

    # elements_buffer = _cl.Buffer(
    #     ctx,
    #     mf.READ_ONLY | mf.COPY_HOST_PTR,
    #     hostbuf=grid.elements.ravel(order="F"),
    # )

    points_buffer = _cl.Buffer(
        ctx,
        mf.READ_ONLY | mf.COPY_HOST_PTR,
        hostbuf=local_points.ravel(order="F"),
    )

    neighbor_indices_buffer = _cl.Buffer(
        ctx,
        mf.READ_ONLY | mf.COPY_HOST_PTR,
        hostbuf=grid.element_neighbors.indices,
    )

    neighbor_indexptr_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=grid.element_neighbors.indexptr)

    coefficients_buffer = _cl.Buffer(ctx, mf.READ_ONLY, size=result_type.itemsize * ncoeffs)

    result_buffer = _cl.Buffer(ctx, mf.READ_WRITE, size=4 * result_type.itemsize * ncoeffs)

    if len(kernel_parameters) == 0:
        kernel_parameters = [0]

    kernel_parameters_buffer = _cl.Buffer(
        ctx,
        mf.READ_ONLY | mf.COPY_HOST_PTR,
        hostbuf=_np.array(kernel_parameters, dtype="float64"),
    )

    options = {"MAX_POINTS": max_nneighbors * npoints, "NPOINTS": npoints}
    if result_type == "complex128":
        options["COMPLEX_KERNEL"] = None

    kernel_name = "near_field_evaluator_" + mode
    kernel = get_kernel_from_name(kernel_name, options)

    def evaluator(coeffs):
        """Actually evaluate the near-field correction."""
        result = _np.empty(4 * ncoeffs, dtype=result_type)
        with bempp_cl.api.Timer(message="Singular Corrections Evaluator"):
            with _cl.CommandQueue(ctx, device=device) as queue:
                _cl.enqueue_copy(queue, coefficients_buffer, coeffs.astype(result_type))
                _cl.enqueue_fill_buffer(
                    queue,
                    result_buffer,
                    _np.uint8(0),
                    0,
                    result_type.itemsize * ncoeffs,
                )
                kernel(
                    queue,
                    (grid.number_of_elements,),
                    (1,),
                    grid_buffer,
                    neighbor_indices_buffer,
                    neighbor_indexptr_buffer,
                    points_buffer,
                    coefficients_buffer,
                    result_buffer,
                    kernel_parameters_buffer,
                    _np.uint32(grid.number_of_elements),
                )
                _cl.enqueue_copy(queue, result, result_buffer)

        return result

    return evaluator


def debug_fmm(targets, sources, charges, mode, kernel_parameters, fmm_result):
    """Compare the result of an FMM result with the corresponding dense computation."""
    import bempp_cl.api

    dense_result = dense_interaction_evaluator(targets, sources, charges, mode, kernel_parameters).reshape(-1, 4)

    rel_error = _np.max(_np.abs(dense_result - fmm_result) / _np.abs(fmm_result))

    bempp_cl.api.log(f"FMM error: {rel_error}.")

    return dense_result


def dense_interaction_evaluator(targets, sources, charges, mode, kernel_parameters):
    """
    Dense evaluation of interaction between sources and targets.

    Parameters
    ----------
    targets : ndarray
        M x 3 array of target points.
    sources : ndarray
        N x 3 array of source points.
    charges : ndarray
        N array of charges.
    mode : string
        Either 'laplace', 'helmholtz', 'modified_helmholtz'
    kernel_parameters : ndarray
        Array with kernel parameters
    kernel_type : dtype
        Type of the kernel (numpy.float64 or numpy.complex128)

    Returns the dense evaluation of the interaction between sources
    and targets with the given charges.
    """
    if mode == "laplace":
        kernel = laplace_kernel
        kernel_type = _np.float64
    elif mode == "helmholtz":
        kernel = helmholtz_kernel
        kernel_type = _np.complex128
    elif mode == "modified_helmholtz":
        kernel = modified_helmholtz_kernel
        kernel_type = _np.float64
    else:
        raise ValueError("Unknown value for 'kernel_function'.")

    return dense_interaction_evaluator_impl(targets, sources, charges, kernel, kernel_parameters, kernel_type).reshape(
        -1, 4
    )


@_numba.jit(nopython=True, parallel=True, error_model="numpy", fastmath=True, boundscheck=False)
def dense_interaction_evaluator_impl(targets, sources, charges, kernel, kernel_parameters, kernel_type):
    """
    Dense evaluation of interaction between sources and targets.

    Parameters
    ----------
    targets : ndarray
        M x 3 array of target points.
    sources : ndarray
        N x 3 array of source points.
    charges : ndarray
        N array of charges.
    kernel : Numba function object
        The kernel object (either helpers.laplace,
        helpers.helmholtz or helpers.modified_helmholtz)
    kernel_parameters : ndarray
        Array with kernel parameters
    kernel_type : dtype
        Type of the kernel (numpy.float64 or numpy.complex128)

    Returns the dense evaluation of the interaction between sources
    and targets with the given charges.
    """
    dtype = sources.dtype

    sources = sources.T.copy()
    targets = targets.T.copy()

    ntargets = targets.shape[1]
    nsources = sources.shape[1]
    result = _np.zeros(4 * ntargets, dtype=kernel_type)

    for target_index in _numba.prange(ntargets):
        current_target = targets[:, target_index].copy().reshape((3, 1))
        vals = kernel(current_target, sources, kernel_parameters, dtype, kernel_type)
        for source_index in range(nsources):
            for local_index in range(4):
                result[4 * target_index + local_index] += vals[4 * source_index + local_index] * charges[source_index]

    return result



import numpy as np

## general gain calculation
def dB(x):
    """
    Convert a value to decibels (dB).

    Parameters
    ----------
    x : float or numpy array
        Value(s) to be converted to decibels.

    Returns
    -------
    dB_value : float or numpy array
        Value(s) converted to decibels (dB).

    Notes
    -----
    This function takes a numeric value or an array of values 'x' and converts
    them to their equivalent decibel (dB) representation using the formula
    20 * log10(abs(x)). The resulting dB value(s) provide a logarithmic
    representation of the input value(s) suitable for expressing ratios
    and differences in magnitude.
    """
    return 20*np.log10(np.abs(x))

def dB_10log(x):
    """
    Convert a value to decibels using the base 10 logarithm.

    Parameters
    ----------
    x : float or numpy array
        Value(s) to be converted to decibels using the base 10 logarithm.

    Returns
    -------
    dB_value : float or numpy array
        Value(s) converted to decibels using the base 10 logarithm.

    Notes
    -----
    This function takes a numeric value or an array of values 'x' and converts
    them to their equivalent decibel (dB) representation using the base 10
    logarithm, i.e., 10 * log10(x). The resulting dB value(s) provide a
    logarithmic representation of the input value(s) suitable for expressing
    ratios and differences in magnitude.
    """
    return 10 * np.log10(x)

def dB_ref(x, freq, ref):
    """
    Estimate the gain in decibels relative to a reference frequency.

    Parameters
    ----------
    x : numpy array
        Array of values to be normalized and converted to decibels.

    freq : numpy array
        Frequency axis corresponding to the values in 'x'.

    ref : float
        Reference frequency (Hz) to which the gain will be normalized.

    Returns
    -------
    gain_dB : numpy array
        Gain values in decibels relative to the reference frequency.

    Notes
    -----
    This function estimates the gain in decibels of the input 'x' array with respect to
    a reference frequency 'ref'. The input 'freq' should be the frequency axis corresponding
    to the values in 'x'. The gain is calculated by normalizing the input values to their
    maximum value at the reference frequency and then converting them to decibels.
    The resulting gain_dB values represent the gain of 'x' relative to the reference frequency.
    """
    indref = int((np.abs(freq - ref)).argmin())
    return 20*np.log10(np.abs(x)/np.max(np.abs(x[indref])))

def dB_zero(x):
    """
    Convert values to decibels relative to the maximum value.

    Parameters
    ----------
    x : numpy array
        Array of values to be normalized and converted to decibels.

    Returns
    -------
    gain_dB : numpy array
        Gain values in decibels relative to the maximum value.

    Notes
    -----
    This function converts the input 'x' array of values to decibels relative to its
    maximum value. The gain_dB values represent the relative magnitude of the input values
    expressed in decibels. The function normalizes the input values to their maximum value
    and then converts them to decibels using the formula 20 * log10(abs(x)/max(abs(x))).
    """
    return 20*np.log10(np.abs(x)/np.max(np.abs(x)))

## SPL calculation
def dBSPL(x):
    """
    Convert RMS signal values to dBSPL (decibels Sound Pressure Level).

    Parameters
    ----------
    x : numpy array
        Array of RMS signal values to be converted to dBSPL.

    Returns
    -------
    dBSPL_value : numpy array
        dBSPL values corresponding to the input RMS signal values.

    Notes
    -----
    This function converts the input 'x' array of RMS signal values to their equivalent
    values in dBSPL (decibels Sound Pressure Level). The conversion is performed using the
    formula 20 * log10(abs(x) / 2e-5), where 2e-5 Pa represents the reference sound pressure.
    The resulting dBSPL values provide a logarithmic representation of the RMS signal magnitude
    in decibels relative to the reference sound pressure level.
    """
    return 20*np.log10(np.abs(x)/2e-5)

def SPL(x):
    """
    Convert peak signal values to dBSPL (decibels Sound Pressure Level), RMS.

    Parameters
    ----------
    x : numpy array
        Array of peak signal values to be converted to dBSPL.

    Returns
    -------
    dBSPL_value : numpy array
        dBSPL values corresponding to the input peak signal values.

    Notes
    -----
    This function converts the input 'x' array of peak signal values to their equivalent
    values in dBSPL (decibels Sound Pressure Level). The conversion is performed using the
    formula 20 * log10(abs(x) / (2e-5 * sqrt(2))), where 2e-5 Pa represents the reference
    sound pressure and sqrt(2) is used to convert the peak values to RMS values.
    The resulting dBSPL values provide a logarithmic representation of the peak signal magnitude
    in decibels relative to the reference sound pressure level.
    """
    return 20*np.log10(np.abs(x)/2e-5/np.sqrt(2))

